/* Generated by Cython 0.28.5 */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_28_5"
#define CYTHON_FUTURE_DIVISION 0
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (0 && PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0; // PyThread_create_key reports success always
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif // TSS (Thread Specific Storage) API
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
  #define PyObject_Unicode             PyObject_Str
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : (Py_INCREF(func), func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__zmq__backend__cython__constants
#define __PYX_HAVE_API__zmq__backend__cython__constants
/* Early includes */
#include "zmq_compat.h"
#include "zmq.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "zmq/backend/cython/constants.pyx",
  "zmq/backend/cython/constants.pxi",
};

/*--- Type declarations ---*/

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* Globals.proto */
static PyObject* __Pyx_Globals(void);

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'zmq.backend.cython.libzmq' */

/* Module declarations from 'zmq.backend.cython.constants' */
#define __Pyx_MODULE_NAME "zmq.backend.cython.constants"
extern int __pyx_module_is_main_zmq__backend__cython__constants;
int __pyx_module_is_main_zmq__backend__cython__constants = 0;

/* Implementation of 'zmq.backend.cython.constants' */
static const char __pyx_k_FD[] = "FD";
static const char __pyx_k_HWM[] = "HWM";
static const char __pyx_k_PUB[] = "PUB";
static const char __pyx_k_REP[] = "REP";
static const char __pyx_k_REQ[] = "REQ";
static const char __pyx_k_SUB[] = "SUB";
static const char __pyx_k_TOS[] = "TOS";
static const char __pyx_k_all[] = "__all__";
static const char __pyx_k_DISH[] = "DISH";
static const char __pyx_k_EFSM[] = "EFSM";
static const char __pyx_k_IPV6[] = "IPV6";
static const char __pyx_k_MORE[] = "MORE";
static const char __pyx_k_NULL[] = "NULL";
static const char __pyx_k_PAIR[] = "PAIR";
static const char __pyx_k_PULL[] = "PULL";
static const char __pyx_k_PUSH[] = "PUSH";
static const char __pyx_k_RATE[] = "RATE";
static const char __pyx_k_SWAP[] = "SWAP";
static const char __pyx_k_TYPE[] = "TYPE";
static const char __pyx_k_XPUB[] = "XPUB";
static const char __pyx_k_XREP[] = "XREP";
static const char __pyx_k_XREQ[] = "XREQ";
static const char __pyx_k_XSUB[] = "XSUB";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_CURVE[] = "CURVE";
static const char __pyx_k_DGRAM[] = "DGRAM";
static const char __pyx_k_ETERM[] = "ETERM";
static const char __pyx_k_PLAIN[] = "PLAIN";
static const char __pyx_k_QUEUE[] = "QUEUE";
static const char __pyx_k_RADIO[] = "RADIO";
static const char __pyx_k_SRCFD[] = "SRCFD";
static const char __pyx_k_BLOCKY[] = "BLOCKY";
static const char __pyx_k_CLIENT[] = "CLIENT";
static const char __pyx_k_DEALER[] = "DEALER";
static const char __pyx_k_EAGAIN[] = "EAGAIN";
static const char __pyx_k_EFAULT[] = "EFAULT";
static const char __pyx_k_EINVAL[] = "EINVAL";
static const char __pyx_k_ENODEV[] = "ENODEV";
static const char __pyx_k_ENOMEM[] = "ENOMEM";
static const char __pyx_k_EVENTS[] = "EVENTS";
static const char __pyx_k_GATHER[] = "GATHER";
static const char __pyx_k_GSSAPI[] = "GSSAPI";
static const char __pyx_k_LINGER[] = "LINGER";
static const char __pyx_k_POLLIN[] = "POLLIN";
static const char __pyx_k_RCVBUF[] = "RCVBUF";
static const char __pyx_k_RCVHWM[] = "RCVHWM";
static const char __pyx_k_ROUTER[] = "ROUTER";
static const char __pyx_k_SERVER[] = "SERVER";
static const char __pyx_k_SHARED[] = "SHARED";
static const char __pyx_k_SNDBUF[] = "SNDBUF";
static const char __pyx_k_SNDHWM[] = "SNDHWM";
static const char __pyx_k_STREAM[] = "STREAM";
static const char __pyx_k_USE_FD[] = "USE_FD";
static const char __pyx_k_BACKLOG[] = "BACKLOG";
static const char __pyx_k_ENOBUFS[] = "ENOBUFS";
static const char __pyx_k_ENOTSUP[] = "ENOTSUP";
static const char __pyx_k_NOBLOCK[] = "NOBLOCK";
static const char __pyx_k_POLLERR[] = "POLLERR";
static const char __pyx_k_POLLOUT[] = "POLLOUT";
static const char __pyx_k_POLLPRI[] = "POLLPRI";
static const char __pyx_k_RCVMORE[] = "RCVMORE";
static const char __pyx_k_SCATTER[] = "SCATTER";
static const char __pyx_k_SNDMORE[] = "SNDMORE";
static const char __pyx_k_VERSION[] = "VERSION";
static const char __pyx_k_AFFINITY[] = "AFFINITY";
static const char __pyx_k_CONFLATE[] = "CONFLATE";
static const char __pyx_k_DONTWAIT[] = "DONTWAIT";
static const char __pyx_k_EMSGSIZE[] = "EMSGSIZE";
static const char __pyx_k_EMTHREAD[] = "EMTHREAD";
static const char __pyx_k_ENETDOWN[] = "ENETDOWN";
static const char __pyx_k_ENOTCONN[] = "ENOTCONN";
static const char __pyx_k_ENOTSOCK[] = "ENOTSOCK";
static const char __pyx_k_IDENTITY[] = "IDENTITY";
static const char __pyx_k_IPV4ONLY[] = "IPV4ONLY";
static const char __pyx_k_RCVTIMEO[] = "RCVTIMEO";
static const char __pyx_k_SNDTIMEO[] = "SNDTIMEO";
static const char __pyx_k_STREAMER[] = "STREAMER";
static const char __pyx_k_UPSTREAM[] = "UPSTREAM";
static const char __pyx_k_DRAFT_API[] = "DRAFT_API";
static const char __pyx_k_ENETRESET[] = "ENETRESET";
static const char __pyx_k_ETIMEDOUT[] = "ETIMEDOUT";
static const char __pyx_k_EVENT_ALL[] = "EVENT_ALL";
static const char __pyx_k_FORWARDER[] = "FORWARDER";
static const char __pyx_k_IMMEDIATE[] = "IMMEDIATE";
static const char __pyx_k_MECHANISM[] = "MECHANISM";
static const char __pyx_k_SUBSCRIBE[] = "SUBSCRIBE";
static const char __pyx_k_TCP_MAXRT[] = "TCP_MAXRT";
static const char __pyx_k_DOWNSTREAM[] = "DOWNSTREAM";
static const char __pyx_k_EADDRINUSE[] = "EADDRINUSE";
static const char __pyx_k_ECONNRESET[] = "ECONNRESET";
static const char __pyx_k_HAUSNUMERO[] = "HAUSNUMERO";
static const char __pyx_k_IO_THREADS[] = "IO_THREADS";
static const char __pyx_k_MAXMSGSIZE[] = "MAXMSGSIZE";
static const char __pyx_k_MCAST_LOOP[] = "MCAST_LOOP";
static const char __pyx_k_ROUTER_RAW[] = "ROUTER_RAW";
static const char __pyx_k_ZAP_DOMAIN[] = "ZAP_DOMAIN";
static const char __pyx_k_CONNECT_RID[] = "CONNECT_RID";
static const char __pyx_k_EINPROGRESS[] = "EINPROGRESS";
static const char __pyx_k_ENETUNREACH[] = "ENETUNREACH";
static const char __pyx_k_MAX_SOCKETS[] = "MAX_SOCKETS";
static const char __pyx_k_REQ_RELAXED[] = "REQ_RELAXED";
static const char __pyx_k_SOCKS_PROXY[] = "SOCKS_PROXY";
static const char __pyx_k_THREAD_SAFE[] = "THREAD_SAFE";
static const char __pyx_k_UNSUBSCRIBE[] = "UNSUBSCRIBE";
static const char __pyx_k_XPUB_MANUAL[] = "XPUB_MANUAL";
static const char __pyx_k_XPUB_NODROP[] = "XPUB_NODROP";
static const char __pyx_k_CURVE_SERVER[] = "CURVE_SERVER";
static const char __pyx_k_EAFNOSUPPORT[] = "EAFNOSUPPORT";
static const char __pyx_k_ECONNABORTED[] = "ECONNABORTED";
static const char __pyx_k_ECONNREFUSED[] = "ECONNREFUSED";
static const char __pyx_k_EHOSTUNREACH[] = "EHOSTUNREACH";
static const char __pyx_k_EVENT_CLOSED[] = "EVENT_CLOSED";
static const char __pyx_k_PLAIN_SERVER[] = "PLAIN_SERVER";
static const char __pyx_k_PROBE_ROUTER[] = "PROBE_ROUTER";
static const char __pyx_k_RECOVERY_IVL[] = "RECOVERY_IVL";
static const char __pyx_k_SOCKET_LIMIT[] = "SOCKET_LIMIT";
static const char __pyx_k_XPUB_VERBOSE[] = "XPUB_VERBOSE";
static const char __pyx_k_0MQ_Constants[] = "0MQ Constants.";
static const char __pyx_k_EADDRNOTAVAIL[] = "EADDRNOTAVAIL";
static const char __pyx_k_GSSAPI_SERVER[] = "GSSAPI_SERVER";
static const char __pyx_k_HANDSHAKE_IVL[] = "HANDSHAKE_IVL";
static const char __pyx_k_HEARTBEAT_IVL[] = "HEARTBEAT_IVL";
static const char __pyx_k_HEARTBEAT_TTL[] = "HEARTBEAT_TTL";
static const char __pyx_k_LAST_ENDPOINT[] = "LAST_ENDPOINT";
static const char __pyx_k_RECONNECT_IVL[] = "RECONNECT_IVL";
static const char __pyx_k_REQ_CORRELATE[] = "REQ_CORRELATE";
static const char __pyx_k_STREAM_NOTIFY[] = "STREAM_NOTIFY";
static const char __pyx_k_TCP_KEEPALIVE[] = "TCP_KEEPALIVE";
static const char __pyx_k_VERSION_MAJOR[] = "VERSION_MAJOR";
static const char __pyx_k_VERSION_MINOR[] = "VERSION_MINOR";
static const char __pyx_k_VERSION_PATCH[] = "VERSION_PATCH";
static const char __pyx_k_XPUB_VERBOSER[] = "XPUB_VERBOSER";
static const char __pyx_k_ENOCOMPATPROTO[] = "ENOCOMPATPROTO";
static const char __pyx_k_EVENT_ACCEPTED[] = "EVENT_ACCEPTED";
static const char __pyx_k_IPC_FILTER_GID[] = "IPC_FILTER_GID";
static const char __pyx_k_IPC_FILTER_PID[] = "IPC_FILTER_PID";
static const char __pyx_k_IPC_FILTER_UID[] = "IPC_FILTER_UID";
static const char __pyx_k_MULTICAST_HOPS[] = "MULTICAST_HOPS";
static const char __pyx_k_PLAIN_PASSWORD[] = "PLAIN_PASSWORD";
static const char __pyx_k_PLAIN_USERNAME[] = "PLAIN_USERNAME";
static const char __pyx_k_POLLITEMS_DFLT[] = "POLLITEMS_DFLT";
static const char __pyx_k_CONNECT_TIMEOUT[] = "CONNECT_TIMEOUT";
static const char __pyx_k_CURVE_PUBLICKEY[] = "CURVE_PUBLICKEY";
static const char __pyx_k_CURVE_SECRETKEY[] = "CURVE_SECRETKEY";
static const char __pyx_k_CURVE_SERVERKEY[] = "CURVE_SERVERKEY";
static const char __pyx_k_EPROTONOSUPPORT[] = "EPROTONOSUPPORT";
static const char __pyx_k_EVENT_CONNECTED[] = "EVENT_CONNECTED";
static const char __pyx_k_EVENT_LISTENING[] = "EVENT_LISTENING";
static const char __pyx_k_FAIL_UNROUTABLE[] = "FAIL_UNROUTABLE";
static const char __pyx_k_INVERT_MATCHING[] = "INVERT_MATCHING";
static const char __pyx_k_IO_THREADS_DFLT[] = "IO_THREADS_DFLT";
static const char __pyx_k_ROUTER_BEHAVIOR[] = "ROUTER_BEHAVIOR";
static const char __pyx_k_ROUTER_HANDOVER[] = "ROUTER_HANDOVER";
static const char __pyx_k_THREAD_PRIORITY[] = "THREAD_PRIORITY";
static const char __pyx_k_GSSAPI_PLAINTEXT[] = "GSSAPI_PLAINTEXT";
static const char __pyx_k_GSSAPI_PRINCIPAL[] = "GSSAPI_PRINCIPAL";
static const char __pyx_k_MAX_SOCKETS_DFLT[] = "MAX_SOCKETS_DFLT";
static const char __pyx_k_ROUTER_MANDATORY[] = "ROUTER_MANDATORY";
static const char __pyx_k_VMCI_BUFFER_SIZE[] = "VMCI_BUFFER_SIZE";
static const char __pyx_k_XPUB_WELCOME_MSG[] = "XPUB_WELCOME_MSG";
static const char __pyx_k_EVENT_BIND_FAILED[] = "EVENT_BIND_FAILED";
static const char __pyx_k_HEARTBEAT_TIMEOUT[] = "HEARTBEAT_TIMEOUT";
static const char __pyx_k_MULTICAST_MAXTPDU[] = "MULTICAST_MAXTPDU";
static const char __pyx_k_RECONNECT_IVL_MAX[] = "RECONNECT_IVL_MAX";
static const char __pyx_k_RECOVERY_IVL_MSEC[] = "RECOVERY_IVL_MSEC";
static const char __pyx_k_TCP_ACCEPT_FILTER[] = "TCP_ACCEPT_FILTER";
static const char __pyx_k_TCP_KEEPALIVE_CNT[] = "TCP_KEEPALIVE_CNT";
static const char __pyx_k_EVENT_CLOSE_FAILED[] = "EVENT_CLOSE_FAILED";
static const char __pyx_k_EVENT_DISCONNECTED[] = "EVENT_DISCONNECTED";
static const char __pyx_k_TCP_KEEPALIVE_IDLE[] = "TCP_KEEPALIVE_IDLE";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_EVENT_ACCEPT_FAILED[] = "EVENT_ACCEPT_FAILED";
static const char __pyx_k_TCP_KEEPALIVE_INTVL[] = "TCP_KEEPALIVE_INTVL";
static const char __pyx_k_THREAD_SCHED_POLICY[] = "THREAD_SCHED_POLICY";
static const char __pyx_k_THREAD_PRIORITY_DFLT[] = "THREAD_PRIORITY_DFLT";
static const char __pyx_k_VMCI_BUFFER_MAX_SIZE[] = "VMCI_BUFFER_MAX_SIZE";
static const char __pyx_k_VMCI_BUFFER_MIN_SIZE[] = "VMCI_BUFFER_MIN_SIZE";
static const char __pyx_k_VMCI_CONNECT_TIMEOUT[] = "VMCI_CONNECT_TIMEOUT";
static const char __pyx_k_EVENT_CONNECT_DELAYED[] = "EVENT_CONNECT_DELAYED";
static const char __pyx_k_EVENT_CONNECT_RETRIED[] = "EVENT_CONNECT_RETRIED";
static const char __pyx_k_EVENT_MONITOR_STOPPED[] = "EVENT_MONITOR_STOPPED";
static const char __pyx_k_DELAY_ATTACH_ON_CONNECT[] = "DELAY_ATTACH_ON_CONNECT";
static const char __pyx_k_GSSAPI_SERVICE_PRINCIPAL[] = "GSSAPI_SERVICE_PRINCIPAL";
static const char __pyx_k_THREAD_SCHED_POLICY_DFLT[] = "THREAD_SCHED_POLICY_DFLT";
static PyObject *__pyx_n_s_AFFINITY;
static PyObject *__pyx_n_s_BACKLOG;
static PyObject *__pyx_n_s_BLOCKY;
static PyObject *__pyx_n_s_CLIENT;
static PyObject *__pyx_n_s_CONFLATE;
static PyObject *__pyx_n_s_CONNECT_RID;
static PyObject *__pyx_n_s_CONNECT_TIMEOUT;
static PyObject *__pyx_n_s_CURVE;
static PyObject *__pyx_n_s_CURVE_PUBLICKEY;
static PyObject *__pyx_n_s_CURVE_SECRETKEY;
static PyObject *__pyx_n_s_CURVE_SERVER;
static PyObject *__pyx_n_s_CURVE_SERVERKEY;
static PyObject *__pyx_n_s_DEALER;
static PyObject *__pyx_n_s_DELAY_ATTACH_ON_CONNECT;
static PyObject *__pyx_n_s_DGRAM;
static PyObject *__pyx_n_s_DISH;
static PyObject *__pyx_n_s_DONTWAIT;
static PyObject *__pyx_n_s_DOWNSTREAM;
static PyObject *__pyx_n_s_DRAFT_API;
static PyObject *__pyx_n_s_EADDRINUSE;
static PyObject *__pyx_n_s_EADDRNOTAVAIL;
static PyObject *__pyx_n_s_EAFNOSUPPORT;
static PyObject *__pyx_n_s_EAGAIN;
static PyObject *__pyx_n_s_ECONNABORTED;
static PyObject *__pyx_n_s_ECONNREFUSED;
static PyObject *__pyx_n_s_ECONNRESET;
static PyObject *__pyx_n_s_EFAULT;
static PyObject *__pyx_n_s_EFSM;
static PyObject *__pyx_n_s_EHOSTUNREACH;
static PyObject *__pyx_n_s_EINPROGRESS;
static PyObject *__pyx_n_s_EINVAL;
static PyObject *__pyx_n_s_EMSGSIZE;
static PyObject *__pyx_n_s_EMTHREAD;
static PyObject *__pyx_n_s_ENETDOWN;
static PyObject *__pyx_n_s_ENETRESET;
static PyObject *__pyx_n_s_ENETUNREACH;
static PyObject *__pyx_n_s_ENOBUFS;
static PyObject *__pyx_n_s_ENOCOMPATPROTO;
static PyObject *__pyx_n_s_ENODEV;
static PyObject *__pyx_n_s_ENOMEM;
static PyObject *__pyx_n_s_ENOTCONN;
static PyObject *__pyx_n_s_ENOTSOCK;
static PyObject *__pyx_n_s_ENOTSUP;
static PyObject *__pyx_n_s_EPROTONOSUPPORT;
static PyObject *__pyx_n_s_ETERM;
static PyObject *__pyx_n_s_ETIMEDOUT;
static PyObject *__pyx_n_s_EVENTS;
static PyObject *__pyx_n_s_EVENT_ACCEPTED;
static PyObject *__pyx_n_s_EVENT_ACCEPT_FAILED;
static PyObject *__pyx_n_s_EVENT_ALL;
static PyObject *__pyx_n_s_EVENT_BIND_FAILED;
static PyObject *__pyx_n_s_EVENT_CLOSED;
static PyObject *__pyx_n_s_EVENT_CLOSE_FAILED;
static PyObject *__pyx_n_s_EVENT_CONNECTED;
static PyObject *__pyx_n_s_EVENT_CONNECT_DELAYED;
static PyObject *__pyx_n_s_EVENT_CONNECT_RETRIED;
static PyObject *__pyx_n_s_EVENT_DISCONNECTED;
static PyObject *__pyx_n_s_EVENT_LISTENING;
static PyObject *__pyx_n_s_EVENT_MONITOR_STOPPED;
static PyObject *__pyx_n_s_FAIL_UNROUTABLE;
static PyObject *__pyx_n_s_FD;
static PyObject *__pyx_n_s_FORWARDER;
static PyObject *__pyx_n_s_GATHER;
static PyObject *__pyx_n_s_GSSAPI;
static PyObject *__pyx_n_s_GSSAPI_PLAINTEXT;
static PyObject *__pyx_n_s_GSSAPI_PRINCIPAL;
static PyObject *__pyx_n_s_GSSAPI_SERVER;
static PyObject *__pyx_n_s_GSSAPI_SERVICE_PRINCIPAL;
static PyObject *__pyx_n_s_HANDSHAKE_IVL;
static PyObject *__pyx_n_s_HAUSNUMERO;
static PyObject *__pyx_n_s_HEARTBEAT_IVL;
static PyObject *__pyx_n_s_HEARTBEAT_TIMEOUT;
static PyObject *__pyx_n_s_HEARTBEAT_TTL;
static PyObject *__pyx_n_s_HWM;
static PyObject *__pyx_n_s_IDENTITY;
static PyObject *__pyx_n_s_IMMEDIATE;
static PyObject *__pyx_n_s_INVERT_MATCHING;
static PyObject *__pyx_n_s_IO_THREADS;
static PyObject *__pyx_n_s_IO_THREADS_DFLT;
static PyObject *__pyx_n_s_IPC_FILTER_GID;
static PyObject *__pyx_n_s_IPC_FILTER_PID;
static PyObject *__pyx_n_s_IPC_FILTER_UID;
static PyObject *__pyx_n_s_IPV4ONLY;
static PyObject *__pyx_n_s_IPV6;
static PyObject *__pyx_n_s_LAST_ENDPOINT;
static PyObject *__pyx_n_s_LINGER;
static PyObject *__pyx_n_s_MAXMSGSIZE;
static PyObject *__pyx_n_s_MAX_SOCKETS;
static PyObject *__pyx_n_s_MAX_SOCKETS_DFLT;
static PyObject *__pyx_n_s_MCAST_LOOP;
static PyObject *__pyx_n_s_MECHANISM;
static PyObject *__pyx_n_s_MORE;
static PyObject *__pyx_n_s_MULTICAST_HOPS;
static PyObject *__pyx_n_s_MULTICAST_MAXTPDU;
static PyObject *__pyx_n_s_NOBLOCK;
static PyObject *__pyx_n_s_NULL;
static PyObject *__pyx_n_s_PAIR;
static PyObject *__pyx_n_s_PLAIN;
static PyObject *__pyx_n_s_PLAIN_PASSWORD;
static PyObject *__pyx_n_s_PLAIN_SERVER;
static PyObject *__pyx_n_s_PLAIN_USERNAME;
static PyObject *__pyx_n_s_POLLERR;
static PyObject *__pyx_n_s_POLLIN;
static PyObject *__pyx_n_s_POLLITEMS_DFLT;
static PyObject *__pyx_n_s_POLLOUT;
static PyObject *__pyx_n_s_POLLPRI;
static PyObject *__pyx_n_s_PROBE_ROUTER;
static PyObject *__pyx_n_s_PUB;
static PyObject *__pyx_n_s_PULL;
static PyObject *__pyx_n_s_PUSH;
static PyObject *__pyx_n_s_QUEUE;
static PyObject *__pyx_n_s_RADIO;
static PyObject *__pyx_n_s_RATE;
static PyObject *__pyx_n_s_RCVBUF;
static PyObject *__pyx_n_s_RCVHWM;
static PyObject *__pyx_n_s_RCVMORE;
static PyObject *__pyx_n_s_RCVTIMEO;
static PyObject *__pyx_n_s_RECONNECT_IVL;
static PyObject *__pyx_n_s_RECONNECT_IVL_MAX;
static PyObject *__pyx_n_s_RECOVERY_IVL;
static PyObject *__pyx_n_s_RECOVERY_IVL_MSEC;
static PyObject *__pyx_n_s_REP;
static PyObject *__pyx_n_s_REQ;
static PyObject *__pyx_n_s_REQ_CORRELATE;
static PyObject *__pyx_n_s_REQ_RELAXED;
static PyObject *__pyx_n_s_ROUTER;
static PyObject *__pyx_n_s_ROUTER_BEHAVIOR;
static PyObject *__pyx_n_s_ROUTER_HANDOVER;
static PyObject *__pyx_n_s_ROUTER_MANDATORY;
static PyObject *__pyx_n_s_ROUTER_RAW;
static PyObject *__pyx_n_s_SCATTER;
static PyObject *__pyx_n_s_SERVER;
static PyObject *__pyx_n_s_SHARED;
static PyObject *__pyx_n_s_SNDBUF;
static PyObject *__pyx_n_s_SNDHWM;
static PyObject *__pyx_n_s_SNDMORE;
static PyObject *__pyx_n_s_SNDTIMEO;
static PyObject *__pyx_n_s_SOCKET_LIMIT;
static PyObject *__pyx_n_s_SOCKS_PROXY;
static PyObject *__pyx_n_s_SRCFD;
static PyObject *__pyx_n_s_STREAM;
static PyObject *__pyx_n_s_STREAMER;
static PyObject *__pyx_n_s_STREAM_NOTIFY;
static PyObject *__pyx_n_s_SUB;
static PyObject *__pyx_n_s_SUBSCRIBE;
static PyObject *__pyx_n_s_SWAP;
static PyObject *__pyx_n_s_TCP_ACCEPT_FILTER;
static PyObject *__pyx_n_s_TCP_KEEPALIVE;
static PyObject *__pyx_n_s_TCP_KEEPALIVE_CNT;
static PyObject *__pyx_n_s_TCP_KEEPALIVE_IDLE;
static PyObject *__pyx_n_s_TCP_KEEPALIVE_INTVL;
static PyObject *__pyx_n_s_TCP_MAXRT;
static PyObject *__pyx_n_s_THREAD_PRIORITY;
static PyObject *__pyx_n_s_THREAD_PRIORITY_DFLT;
static PyObject *__pyx_n_s_THREAD_SAFE;
static PyObject *__pyx_n_s_THREAD_SCHED_POLICY;
static PyObject *__pyx_n_s_THREAD_SCHED_POLICY_DFLT;
static PyObject *__pyx_n_s_TOS;
static PyObject *__pyx_n_s_TYPE;
static PyObject *__pyx_n_s_UNSUBSCRIBE;
static PyObject *__pyx_n_s_UPSTREAM;
static PyObject *__pyx_n_s_USE_FD;
static PyObject *__pyx_n_s_VERSION;
static PyObject *__pyx_n_s_VERSION_MAJOR;
static PyObject *__pyx_n_s_VERSION_MINOR;
static PyObject *__pyx_n_s_VERSION_PATCH;
static PyObject *__pyx_n_s_VMCI_BUFFER_MAX_SIZE;
static PyObject *__pyx_n_s_VMCI_BUFFER_MIN_SIZE;
static PyObject *__pyx_n_s_VMCI_BUFFER_SIZE;
static PyObject *__pyx_n_s_VMCI_CONNECT_TIMEOUT;
static PyObject *__pyx_n_s_XPUB;
static PyObject *__pyx_n_s_XPUB_MANUAL;
static PyObject *__pyx_n_s_XPUB_NODROP;
static PyObject *__pyx_n_s_XPUB_VERBOSE;
static PyObject *__pyx_n_s_XPUB_VERBOSER;
static PyObject *__pyx_n_s_XPUB_WELCOME_MSG;
static PyObject *__pyx_n_s_XREP;
static PyObject *__pyx_n_s_XREQ;
static PyObject *__pyx_n_s_XSUB;
static PyObject *__pyx_n_s_ZAP_DOMAIN;
static PyObject *__pyx_n_s_all;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_test;
/* Late includes */

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_constants(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_constants},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "constants",
    __pyx_k_0MQ_Constants, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_AFFINITY, __pyx_k_AFFINITY, sizeof(__pyx_k_AFFINITY), 0, 0, 1, 1},
  {&__pyx_n_s_BACKLOG, __pyx_k_BACKLOG, sizeof(__pyx_k_BACKLOG), 0, 0, 1, 1},
  {&__pyx_n_s_BLOCKY, __pyx_k_BLOCKY, sizeof(__pyx_k_BLOCKY), 0, 0, 1, 1},
  {&__pyx_n_s_CLIENT, __pyx_k_CLIENT, sizeof(__pyx_k_CLIENT), 0, 0, 1, 1},
  {&__pyx_n_s_CONFLATE, __pyx_k_CONFLATE, sizeof(__pyx_k_CONFLATE), 0, 0, 1, 1},
  {&__pyx_n_s_CONNECT_RID, __pyx_k_CONNECT_RID, sizeof(__pyx_k_CONNECT_RID), 0, 0, 1, 1},
  {&__pyx_n_s_CONNECT_TIMEOUT, __pyx_k_CONNECT_TIMEOUT, sizeof(__pyx_k_CONNECT_TIMEOUT), 0, 0, 1, 1},
  {&__pyx_n_s_CURVE, __pyx_k_CURVE, sizeof(__pyx_k_CURVE), 0, 0, 1, 1},
  {&__pyx_n_s_CURVE_PUBLICKEY, __pyx_k_CURVE_PUBLICKEY, sizeof(__pyx_k_CURVE_PUBLICKEY), 0, 0, 1, 1},
  {&__pyx_n_s_CURVE_SECRETKEY, __pyx_k_CURVE_SECRETKEY, sizeof(__pyx_k_CURVE_SECRETKEY), 0, 0, 1, 1},
  {&__pyx_n_s_CURVE_SERVER, __pyx_k_CURVE_SERVER, sizeof(__pyx_k_CURVE_SERVER), 0, 0, 1, 1},
  {&__pyx_n_s_CURVE_SERVERKEY, __pyx_k_CURVE_SERVERKEY, sizeof(__pyx_k_CURVE_SERVERKEY), 0, 0, 1, 1},
  {&__pyx_n_s_DEALER, __pyx_k_DEALER, sizeof(__pyx_k_DEALER), 0, 0, 1, 1},
  {&__pyx_n_s_DELAY_ATTACH_ON_CONNECT, __pyx_k_DELAY_ATTACH_ON_CONNECT, sizeof(__pyx_k_DELAY_ATTACH_ON_CONNECT), 0, 0, 1, 1},
  {&__pyx_n_s_DGRAM, __pyx_k_DGRAM, sizeof(__pyx_k_DGRAM), 0, 0, 1, 1},
  {&__pyx_n_s_DISH, __pyx_k_DISH, sizeof(__pyx_k_DISH), 0, 0, 1, 1},
  {&__pyx_n_s_DONTWAIT, __pyx_k_DONTWAIT, sizeof(__pyx_k_DONTWAIT), 0, 0, 1, 1},
  {&__pyx_n_s_DOWNSTREAM, __pyx_k_DOWNSTREAM, sizeof(__pyx_k_DOWNSTREAM), 0, 0, 1, 1},
  {&__pyx_n_s_DRAFT_API, __pyx_k_DRAFT_API, sizeof(__pyx_k_DRAFT_API), 0, 0, 1, 1},
  {&__pyx_n_s_EADDRINUSE, __pyx_k_EADDRINUSE, sizeof(__pyx_k_EADDRINUSE), 0, 0, 1, 1},
  {&__pyx_n_s_EADDRNOTAVAIL, __pyx_k_EADDRNOTAVAIL, sizeof(__pyx_k_EADDRNOTAVAIL), 0, 0, 1, 1},
  {&__pyx_n_s_EAFNOSUPPORT, __pyx_k_EAFNOSUPPORT, sizeof(__pyx_k_EAFNOSUPPORT), 0, 0, 1, 1},
  {&__pyx_n_s_EAGAIN, __pyx_k_EAGAIN, sizeof(__pyx_k_EAGAIN), 0, 0, 1, 1},
  {&__pyx_n_s_ECONNABORTED, __pyx_k_ECONNABORTED, sizeof(__pyx_k_ECONNABORTED), 0, 0, 1, 1},
  {&__pyx_n_s_ECONNREFUSED, __pyx_k_ECONNREFUSED, sizeof(__pyx_k_ECONNREFUSED), 0, 0, 1, 1},
  {&__pyx_n_s_ECONNRESET, __pyx_k_ECONNRESET, sizeof(__pyx_k_ECONNRESET), 0, 0, 1, 1},
  {&__pyx_n_s_EFAULT, __pyx_k_EFAULT, sizeof(__pyx_k_EFAULT), 0, 0, 1, 1},
  {&__pyx_n_s_EFSM, __pyx_k_EFSM, sizeof(__pyx_k_EFSM), 0, 0, 1, 1},
  {&__pyx_n_s_EHOSTUNREACH, __pyx_k_EHOSTUNREACH, sizeof(__pyx_k_EHOSTUNREACH), 0, 0, 1, 1},
  {&__pyx_n_s_EINPROGRESS, __pyx_k_EINPROGRESS, sizeof(__pyx_k_EINPROGRESS), 0, 0, 1, 1},
  {&__pyx_n_s_EINVAL, __pyx_k_EINVAL, sizeof(__pyx_k_EINVAL), 0, 0, 1, 1},
  {&__pyx_n_s_EMSGSIZE, __pyx_k_EMSGSIZE, sizeof(__pyx_k_EMSGSIZE), 0, 0, 1, 1},
  {&__pyx_n_s_EMTHREAD, __pyx_k_EMTHREAD, sizeof(__pyx_k_EMTHREAD), 0, 0, 1, 1},
  {&__pyx_n_s_ENETDOWN, __pyx_k_ENETDOWN, sizeof(__pyx_k_ENETDOWN), 0, 0, 1, 1},
  {&__pyx_n_s_ENETRESET, __pyx_k_ENETRESET, sizeof(__pyx_k_ENETRESET), 0, 0, 1, 1},
  {&__pyx_n_s_ENETUNREACH, __pyx_k_ENETUNREACH, sizeof(__pyx_k_ENETUNREACH), 0, 0, 1, 1},
  {&__pyx_n_s_ENOBUFS, __pyx_k_ENOBUFS, sizeof(__pyx_k_ENOBUFS), 0, 0, 1, 1},
  {&__pyx_n_s_ENOCOMPATPROTO, __pyx_k_ENOCOMPATPROTO, sizeof(__pyx_k_ENOCOMPATPROTO), 0, 0, 1, 1},
  {&__pyx_n_s_ENODEV, __pyx_k_ENODEV, sizeof(__pyx_k_ENODEV), 0, 0, 1, 1},
  {&__pyx_n_s_ENOMEM, __pyx_k_ENOMEM, sizeof(__pyx_k_ENOMEM), 0, 0, 1, 1},
  {&__pyx_n_s_ENOTCONN, __pyx_k_ENOTCONN, sizeof(__pyx_k_ENOTCONN), 0, 0, 1, 1},
  {&__pyx_n_s_ENOTSOCK, __pyx_k_ENOTSOCK, sizeof(__pyx_k_ENOTSOCK), 0, 0, 1, 1},
  {&__pyx_n_s_ENOTSUP, __pyx_k_ENOTSUP, sizeof(__pyx_k_ENOTSUP), 0, 0, 1, 1},
  {&__pyx_n_s_EPROTONOSUPPORT, __pyx_k_EPROTONOSUPPORT, sizeof(__pyx_k_EPROTONOSUPPORT), 0, 0, 1, 1},
  {&__pyx_n_s_ETERM, __pyx_k_ETERM, sizeof(__pyx_k_ETERM), 0, 0, 1, 1},
  {&__pyx_n_s_ETIMEDOUT, __pyx_k_ETIMEDOUT, sizeof(__pyx_k_ETIMEDOUT), 0, 0, 1, 1},
  {&__pyx_n_s_EVENTS, __pyx_k_EVENTS, sizeof(__pyx_k_EVENTS), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_ACCEPTED, __pyx_k_EVENT_ACCEPTED, sizeof(__pyx_k_EVENT_ACCEPTED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_ACCEPT_FAILED, __pyx_k_EVENT_ACCEPT_FAILED, sizeof(__pyx_k_EVENT_ACCEPT_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_ALL, __pyx_k_EVENT_ALL, sizeof(__pyx_k_EVENT_ALL), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_BIND_FAILED, __pyx_k_EVENT_BIND_FAILED, sizeof(__pyx_k_EVENT_BIND_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_CLOSED, __pyx_k_EVENT_CLOSED, sizeof(__pyx_k_EVENT_CLOSED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_CLOSE_FAILED, __pyx_k_EVENT_CLOSE_FAILED, sizeof(__pyx_k_EVENT_CLOSE_FAILED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_CONNECTED, __pyx_k_EVENT_CONNECTED, sizeof(__pyx_k_EVENT_CONNECTED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_CONNECT_DELAYED, __pyx_k_EVENT_CONNECT_DELAYED, sizeof(__pyx_k_EVENT_CONNECT_DELAYED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_CONNECT_RETRIED, __pyx_k_EVENT_CONNECT_RETRIED, sizeof(__pyx_k_EVENT_CONNECT_RETRIED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_DISCONNECTED, __pyx_k_EVENT_DISCONNECTED, sizeof(__pyx_k_EVENT_DISCONNECTED), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_LISTENING, __pyx_k_EVENT_LISTENING, sizeof(__pyx_k_EVENT_LISTENING), 0, 0, 1, 1},
  {&__pyx_n_s_EVENT_MONITOR_STOPPED, __pyx_k_EVENT_MONITOR_STOPPED, sizeof(__pyx_k_EVENT_MONITOR_STOPPED), 0, 0, 1, 1},
  {&__pyx_n_s_FAIL_UNROUTABLE, __pyx_k_FAIL_UNROUTABLE, sizeof(__pyx_k_FAIL_UNROUTABLE), 0, 0, 1, 1},
  {&__pyx_n_s_FD, __pyx_k_FD, sizeof(__pyx_k_FD), 0, 0, 1, 1},
  {&__pyx_n_s_FORWARDER, __pyx_k_FORWARDER, sizeof(__pyx_k_FORWARDER), 0, 0, 1, 1},
  {&__pyx_n_s_GATHER, __pyx_k_GATHER, sizeof(__pyx_k_GATHER), 0, 0, 1, 1},
  {&__pyx_n_s_GSSAPI, __pyx_k_GSSAPI, sizeof(__pyx_k_GSSAPI), 0, 0, 1, 1},
  {&__pyx_n_s_GSSAPI_PLAINTEXT, __pyx_k_GSSAPI_PLAINTEXT, sizeof(__pyx_k_GSSAPI_PLAINTEXT), 0, 0, 1, 1},
  {&__pyx_n_s_GSSAPI_PRINCIPAL, __pyx_k_GSSAPI_PRINCIPAL, sizeof(__pyx_k_GSSAPI_PRINCIPAL), 0, 0, 1, 1},
  {&__pyx_n_s_GSSAPI_SERVER, __pyx_k_GSSAPI_SERVER, sizeof(__pyx_k_GSSAPI_SERVER), 0, 0, 1, 1},
  {&__pyx_n_s_GSSAPI_SERVICE_PRINCIPAL, __pyx_k_GSSAPI_SERVICE_PRINCIPAL, sizeof(__pyx_k_GSSAPI_SERVICE_PRINCIPAL), 0, 0, 1, 1},
  {&__pyx_n_s_HANDSHAKE_IVL, __pyx_k_HANDSHAKE_IVL, sizeof(__pyx_k_HANDSHAKE_IVL), 0, 0, 1, 1},
  {&__pyx_n_s_HAUSNUMERO, __pyx_k_HAUSNUMERO, sizeof(__pyx_k_HAUSNUMERO), 0, 0, 1, 1},
  {&__pyx_n_s_HEARTBEAT_IVL, __pyx_k_HEARTBEAT_IVL, sizeof(__pyx_k_HEARTBEAT_IVL), 0, 0, 1, 1},
  {&__pyx_n_s_HEARTBEAT_TIMEOUT, __pyx_k_HEARTBEAT_TIMEOUT, sizeof(__pyx_k_HEARTBEAT_TIMEOUT), 0, 0, 1, 1},
  {&__pyx_n_s_HEARTBEAT_TTL, __pyx_k_HEARTBEAT_TTL, sizeof(__pyx_k_HEARTBEAT_TTL), 0, 0, 1, 1},
  {&__pyx_n_s_HWM, __pyx_k_HWM, sizeof(__pyx_k_HWM), 0, 0, 1, 1},
  {&__pyx_n_s_IDENTITY, __pyx_k_IDENTITY, sizeof(__pyx_k_IDENTITY), 0, 0, 1, 1},
  {&__pyx_n_s_IMMEDIATE, __pyx_k_IMMEDIATE, sizeof(__pyx_k_IMMEDIATE), 0, 0, 1, 1},
  {&__pyx_n_s_INVERT_MATCHING, __pyx_k_INVERT_MATCHING, sizeof(__pyx_k_INVERT_MATCHING), 0, 0, 1, 1},
  {&__pyx_n_s_IO_THREADS, __pyx_k_IO_THREADS, sizeof(__pyx_k_IO_THREADS), 0, 0, 1, 1},
  {&__pyx_n_s_IO_THREADS_DFLT, __pyx_k_IO_THREADS_DFLT, sizeof(__pyx_k_IO_THREADS_DFLT), 0, 0, 1, 1},
  {&__pyx_n_s_IPC_FILTER_GID, __pyx_k_IPC_FILTER_GID, sizeof(__pyx_k_IPC_FILTER_GID), 0, 0, 1, 1},
  {&__pyx_n_s_IPC_FILTER_PID, __pyx_k_IPC_FILTER_PID, sizeof(__pyx_k_IPC_FILTER_PID), 0, 0, 1, 1},
  {&__pyx_n_s_IPC_FILTER_UID, __pyx_k_IPC_FILTER_UID, sizeof(__pyx_k_IPC_FILTER_UID), 0, 0, 1, 1},
  {&__pyx_n_s_IPV4ONLY, __pyx_k_IPV4ONLY, sizeof(__pyx_k_IPV4ONLY), 0, 0, 1, 1},
  {&__pyx_n_s_IPV6, __pyx_k_IPV6, sizeof(__pyx_k_IPV6), 0, 0, 1, 1},
  {&__pyx_n_s_LAST_ENDPOINT, __pyx_k_LAST_ENDPOINT, sizeof(__pyx_k_LAST_ENDPOINT), 0, 0, 1, 1},
  {&__pyx_n_s_LINGER, __pyx_k_LINGER, sizeof(__pyx_k_LINGER), 0, 0, 1, 1},
  {&__pyx_n_s_MAXMSGSIZE, __pyx_k_MAXMSGSIZE, sizeof(__pyx_k_MAXMSGSIZE), 0, 0, 1, 1},
  {&__pyx_n_s_MAX_SOCKETS, __pyx_k_MAX_SOCKETS, sizeof(__pyx_k_MAX_SOCKETS), 0, 0, 1, 1},
  {&__pyx_n_s_MAX_SOCKETS_DFLT, __pyx_k_MAX_SOCKETS_DFLT, sizeof(__pyx_k_MAX_SOCKETS_DFLT), 0, 0, 1, 1},
  {&__pyx_n_s_MCAST_LOOP, __pyx_k_MCAST_LOOP, sizeof(__pyx_k_MCAST_LOOP), 0, 0, 1, 1},
  {&__pyx_n_s_MECHANISM, __pyx_k_MECHANISM, sizeof(__pyx_k_MECHANISM), 0, 0, 1, 1},
  {&__pyx_n_s_MORE, __pyx_k_MORE, sizeof(__pyx_k_MORE), 0, 0, 1, 1},
  {&__pyx_n_s_MULTICAST_HOPS, __pyx_k_MULTICAST_HOPS, sizeof(__pyx_k_MULTICAST_HOPS), 0, 0, 1, 1},
  {&__pyx_n_s_MULTICAST_MAXTPDU, __pyx_k_MULTICAST_MAXTPDU, sizeof(__pyx_k_MULTICAST_MAXTPDU), 0, 0, 1, 1},
  {&__pyx_n_s_NOBLOCK, __pyx_k_NOBLOCK, sizeof(__pyx_k_NOBLOCK), 0, 0, 1, 1},
  {&__pyx_n_s_NULL, __pyx_k_NULL, sizeof(__pyx_k_NULL), 0, 0, 1, 1},
  {&__pyx_n_s_PAIR, __pyx_k_PAIR, sizeof(__pyx_k_PAIR), 0, 0, 1, 1},
  {&__pyx_n_s_PLAIN, __pyx_k_PLAIN, sizeof(__pyx_k_PLAIN), 0, 0, 1, 1},
  {&__pyx_n_s_PLAIN_PASSWORD, __pyx_k_PLAIN_PASSWORD, sizeof(__pyx_k_PLAIN_PASSWORD), 0, 0, 1, 1},
  {&__pyx_n_s_PLAIN_SERVER, __pyx_k_PLAIN_SERVER, sizeof(__pyx_k_PLAIN_SERVER), 0, 0, 1, 1},
  {&__pyx_n_s_PLAIN_USERNAME, __pyx_k_PLAIN_USERNAME, sizeof(__pyx_k_PLAIN_USERNAME), 0, 0, 1, 1},
  {&__pyx_n_s_POLLERR, __pyx_k_POLLERR, sizeof(__pyx_k_POLLERR), 0, 0, 1, 1},
  {&__pyx_n_s_POLLIN, __pyx_k_POLLIN, sizeof(__pyx_k_POLLIN), 0, 0, 1, 1},
  {&__pyx_n_s_POLLITEMS_DFLT, __pyx_k_POLLITEMS_DFLT, sizeof(__pyx_k_POLLITEMS_DFLT), 0, 0, 1, 1},
  {&__pyx_n_s_POLLOUT, __pyx_k_POLLOUT, sizeof(__pyx_k_POLLOUT), 0, 0, 1, 1},
  {&__pyx_n_s_POLLPRI, __pyx_k_POLLPRI, sizeof(__pyx_k_POLLPRI), 0, 0, 1, 1},
  {&__pyx_n_s_PROBE_ROUTER, __pyx_k_PROBE_ROUTER, sizeof(__pyx_k_PROBE_ROUTER), 0, 0, 1, 1},
  {&__pyx_n_s_PUB, __pyx_k_PUB, sizeof(__pyx_k_PUB), 0, 0, 1, 1},
  {&__pyx_n_s_PULL, __pyx_k_PULL, sizeof(__pyx_k_PULL), 0, 0, 1, 1},
  {&__pyx_n_s_PUSH, __pyx_k_PUSH, sizeof(__pyx_k_PUSH), 0, 0, 1, 1},
  {&__pyx_n_s_QUEUE, __pyx_k_QUEUE, sizeof(__pyx_k_QUEUE), 0, 0, 1, 1},
  {&__pyx_n_s_RADIO, __pyx_k_RADIO, sizeof(__pyx_k_RADIO), 0, 0, 1, 1},
  {&__pyx_n_s_RATE, __pyx_k_RATE, sizeof(__pyx_k_RATE), 0, 0, 1, 1},
  {&__pyx_n_s_RCVBUF, __pyx_k_RCVBUF, sizeof(__pyx_k_RCVBUF), 0, 0, 1, 1},
  {&__pyx_n_s_RCVHWM, __pyx_k_RCVHWM, sizeof(__pyx_k_RCVHWM), 0, 0, 1, 1},
  {&__pyx_n_s_RCVMORE, __pyx_k_RCVMORE, sizeof(__pyx_k_RCVMORE), 0, 0, 1, 1},
  {&__pyx_n_s_RCVTIMEO, __pyx_k_RCVTIMEO, sizeof(__pyx_k_RCVTIMEO), 0, 0, 1, 1},
  {&__pyx_n_s_RECONNECT_IVL, __pyx_k_RECONNECT_IVL, sizeof(__pyx_k_RECONNECT_IVL), 0, 0, 1, 1},
  {&__pyx_n_s_RECONNECT_IVL_MAX, __pyx_k_RECONNECT_IVL_MAX, sizeof(__pyx_k_RECONNECT_IVL_MAX), 0, 0, 1, 1},
  {&__pyx_n_s_RECOVERY_IVL, __pyx_k_RECOVERY_IVL, sizeof(__pyx_k_RECOVERY_IVL), 0, 0, 1, 1},
  {&__pyx_n_s_RECOVERY_IVL_MSEC, __pyx_k_RECOVERY_IVL_MSEC, sizeof(__pyx_k_RECOVERY_IVL_MSEC), 0, 0, 1, 1},
  {&__pyx_n_s_REP, __pyx_k_REP, sizeof(__pyx_k_REP), 0, 0, 1, 1},
  {&__pyx_n_s_REQ, __pyx_k_REQ, sizeof(__pyx_k_REQ), 0, 0, 1, 1},
  {&__pyx_n_s_REQ_CORRELATE, __pyx_k_REQ_CORRELATE, sizeof(__pyx_k_REQ_CORRELATE), 0, 0, 1, 1},
  {&__pyx_n_s_REQ_RELAXED, __pyx_k_REQ_RELAXED, sizeof(__pyx_k_REQ_RELAXED), 0, 0, 1, 1},
  {&__pyx_n_s_ROUTER, __pyx_k_ROUTER, sizeof(__pyx_k_ROUTER), 0, 0, 1, 1},
  {&__pyx_n_s_ROUTER_BEHAVIOR, __pyx_k_ROUTER_BEHAVIOR, sizeof(__pyx_k_ROUTER_BEHAVIOR), 0, 0, 1, 1},
  {&__pyx_n_s_ROUTER_HANDOVER, __pyx_k_ROUTER_HANDOVER, sizeof(__pyx_k_ROUTER_HANDOVER), 0, 0, 1, 1},
  {&__pyx_n_s_ROUTER_MANDATORY, __pyx_k_ROUTER_MANDATORY, sizeof(__pyx_k_ROUTER_MANDATORY), 0, 0, 1, 1},
  {&__pyx_n_s_ROUTER_RAW, __pyx_k_ROUTER_RAW, sizeof(__pyx_k_ROUTER_RAW), 0, 0, 1, 1},
  {&__pyx_n_s_SCATTER, __pyx_k_SCATTER, sizeof(__pyx_k_SCATTER), 0, 0, 1, 1},
  {&__pyx_n_s_SERVER, __pyx_k_SERVER, sizeof(__pyx_k_SERVER), 0, 0, 1, 1},
  {&__pyx_n_s_SHARED, __pyx_k_SHARED, sizeof(__pyx_k_SHARED), 0, 0, 1, 1},
  {&__pyx_n_s_SNDBUF, __pyx_k_SNDBUF, sizeof(__pyx_k_SNDBUF), 0, 0, 1, 1},
  {&__pyx_n_s_SNDHWM, __pyx_k_SNDHWM, sizeof(__pyx_k_SNDHWM), 0, 0, 1, 1},
  {&__pyx_n_s_SNDMORE, __pyx_k_SNDMORE, sizeof(__pyx_k_SNDMORE), 0, 0, 1, 1},
  {&__pyx_n_s_SNDTIMEO, __pyx_k_SNDTIMEO, sizeof(__pyx_k_SNDTIMEO), 0, 0, 1, 1},
  {&__pyx_n_s_SOCKET_LIMIT, __pyx_k_SOCKET_LIMIT, sizeof(__pyx_k_SOCKET_LIMIT), 0, 0, 1, 1},
  {&__pyx_n_s_SOCKS_PROXY, __pyx_k_SOCKS_PROXY, sizeof(__pyx_k_SOCKS_PROXY), 0, 0, 1, 1},
  {&__pyx_n_s_SRCFD, __pyx_k_SRCFD, sizeof(__pyx_k_SRCFD), 0, 0, 1, 1},
  {&__pyx_n_s_STREAM, __pyx_k_STREAM, sizeof(__pyx_k_STREAM), 0, 0, 1, 1},
  {&__pyx_n_s_STREAMER, __pyx_k_STREAMER, sizeof(__pyx_k_STREAMER), 0, 0, 1, 1},
  {&__pyx_n_s_STREAM_NOTIFY, __pyx_k_STREAM_NOTIFY, sizeof(__pyx_k_STREAM_NOTIFY), 0, 0, 1, 1},
  {&__pyx_n_s_SUB, __pyx_k_SUB, sizeof(__pyx_k_SUB), 0, 0, 1, 1},
  {&__pyx_n_s_SUBSCRIBE, __pyx_k_SUBSCRIBE, sizeof(__pyx_k_SUBSCRIBE), 0, 0, 1, 1},
  {&__pyx_n_s_SWAP, __pyx_k_SWAP, sizeof(__pyx_k_SWAP), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_ACCEPT_FILTER, __pyx_k_TCP_ACCEPT_FILTER, sizeof(__pyx_k_TCP_ACCEPT_FILTER), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_KEEPALIVE, __pyx_k_TCP_KEEPALIVE, sizeof(__pyx_k_TCP_KEEPALIVE), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_KEEPALIVE_CNT, __pyx_k_TCP_KEEPALIVE_CNT, sizeof(__pyx_k_TCP_KEEPALIVE_CNT), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_KEEPALIVE_IDLE, __pyx_k_TCP_KEEPALIVE_IDLE, sizeof(__pyx_k_TCP_KEEPALIVE_IDLE), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_KEEPALIVE_INTVL, __pyx_k_TCP_KEEPALIVE_INTVL, sizeof(__pyx_k_TCP_KEEPALIVE_INTVL), 0, 0, 1, 1},
  {&__pyx_n_s_TCP_MAXRT, __pyx_k_TCP_MAXRT, sizeof(__pyx_k_TCP_MAXRT), 0, 0, 1, 1},
  {&__pyx_n_s_THREAD_PRIORITY, __pyx_k_THREAD_PRIORITY, sizeof(__pyx_k_THREAD_PRIORITY), 0, 0, 1, 1},
  {&__pyx_n_s_THREAD_PRIORITY_DFLT, __pyx_k_THREAD_PRIORITY_DFLT, sizeof(__pyx_k_THREAD_PRIORITY_DFLT), 0, 0, 1, 1},
  {&__pyx_n_s_THREAD_SAFE, __pyx_k_THREAD_SAFE, sizeof(__pyx_k_THREAD_SAFE), 0, 0, 1, 1},
  {&__pyx_n_s_THREAD_SCHED_POLICY, __pyx_k_THREAD_SCHED_POLICY, sizeof(__pyx_k_THREAD_SCHED_POLICY), 0, 0, 1, 1},
  {&__pyx_n_s_THREAD_SCHED_POLICY_DFLT, __pyx_k_THREAD_SCHED_POLICY_DFLT, sizeof(__pyx_k_THREAD_SCHED_POLICY_DFLT), 0, 0, 1, 1},
  {&__pyx_n_s_TOS, __pyx_k_TOS, sizeof(__pyx_k_TOS), 0, 0, 1, 1},
  {&__pyx_n_s_TYPE, __pyx_k_TYPE, sizeof(__pyx_k_TYPE), 0, 0, 1, 1},
  {&__pyx_n_s_UNSUBSCRIBE, __pyx_k_UNSUBSCRIBE, sizeof(__pyx_k_UNSUBSCRIBE), 0, 0, 1, 1},
  {&__pyx_n_s_UPSTREAM, __pyx_k_UPSTREAM, sizeof(__pyx_k_UPSTREAM), 0, 0, 1, 1},
  {&__pyx_n_s_USE_FD, __pyx_k_USE_FD, sizeof(__pyx_k_USE_FD), 0, 0, 1, 1},
  {&__pyx_n_s_VERSION, __pyx_k_VERSION, sizeof(__pyx_k_VERSION), 0, 0, 1, 1},
  {&__pyx_n_s_VERSION_MAJOR, __pyx_k_VERSION_MAJOR, sizeof(__pyx_k_VERSION_MAJOR), 0, 0, 1, 1},
  {&__pyx_n_s_VERSION_MINOR, __pyx_k_VERSION_MINOR, sizeof(__pyx_k_VERSION_MINOR), 0, 0, 1, 1},
  {&__pyx_n_s_VERSION_PATCH, __pyx_k_VERSION_PATCH, sizeof(__pyx_k_VERSION_PATCH), 0, 0, 1, 1},
  {&__pyx_n_s_VMCI_BUFFER_MAX_SIZE, __pyx_k_VMCI_BUFFER_MAX_SIZE, sizeof(__pyx_k_VMCI_BUFFER_MAX_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_VMCI_BUFFER_MIN_SIZE, __pyx_k_VMCI_BUFFER_MIN_SIZE, sizeof(__pyx_k_VMCI_BUFFER_MIN_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_VMCI_BUFFER_SIZE, __pyx_k_VMCI_BUFFER_SIZE, sizeof(__pyx_k_VMCI_BUFFER_SIZE), 0, 0, 1, 1},
  {&__pyx_n_s_VMCI_CONNECT_TIMEOUT, __pyx_k_VMCI_CONNECT_TIMEOUT, sizeof(__pyx_k_VMCI_CONNECT_TIMEOUT), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB, __pyx_k_XPUB, sizeof(__pyx_k_XPUB), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB_MANUAL, __pyx_k_XPUB_MANUAL, sizeof(__pyx_k_XPUB_MANUAL), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB_NODROP, __pyx_k_XPUB_NODROP, sizeof(__pyx_k_XPUB_NODROP), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB_VERBOSE, __pyx_k_XPUB_VERBOSE, sizeof(__pyx_k_XPUB_VERBOSE), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB_VERBOSER, __pyx_k_XPUB_VERBOSER, sizeof(__pyx_k_XPUB_VERBOSER), 0, 0, 1, 1},
  {&__pyx_n_s_XPUB_WELCOME_MSG, __pyx_k_XPUB_WELCOME_MSG, sizeof(__pyx_k_XPUB_WELCOME_MSG), 0, 0, 1, 1},
  {&__pyx_n_s_XREP, __pyx_k_XREP, sizeof(__pyx_k_XREP), 0, 0, 1, 1},
  {&__pyx_n_s_XREQ, __pyx_k_XREQ, sizeof(__pyx_k_XREQ), 0, 0, 1, 1},
  {&__pyx_n_s_XSUB, __pyx_k_XSUB, sizeof(__pyx_k_XSUB), 0, 0, 1, 1},
  {&__pyx_n_s_ZAP_DOMAIN, __pyx_k_ZAP_DOMAIN, sizeof(__pyx_k_ZAP_DOMAIN), 0, 0, 1, 1},
  {&__pyx_n_s_all, __pyx_k_all, sizeof(__pyx_k_all), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  return 0;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_modinit_global_init_code(void); /*proto*/
static int __Pyx_modinit_variable_export_code(void); /*proto*/
static int __Pyx_modinit_function_export_code(void); /*proto*/
static int __Pyx_modinit_type_init_code(void); /*proto*/
static int __Pyx_modinit_type_import_code(void); /*proto*/
static int __Pyx_modinit_variable_import_code(void); /*proto*/
static int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#if PY_MAJOR_VERSION < 3
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC void
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#else
#ifdef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#endif
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (!(defined(__cplusplus)) || (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ > 4)))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initconstants(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initconstants(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_constants(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_constants(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        result = PyDict_SetItemString(moddict, to_name, value);
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__") < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__") < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static int __pyx_pymod_exec_constants(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m && __pyx_m == __pyx_pyinit_module) return 0;
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_constants(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("constants", __pyx_methods, __pyx_k_0MQ_Constants, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_zmq__backend__cython__constants) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "zmq.backend.cython.constants")) {
      if (unlikely(PyDict_SetItemString(modules, "zmq.backend.cython.constants", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  (void)__Pyx_modinit_type_init_code();
  (void)__Pyx_modinit_type_import_code();
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "zmq/backend/cython/constants.pxi":5
 * #-----------------------------------------------------------------------------
 * 
 * DRAFT_API = PYZMQ_DRAFT_API             # <<<<<<<<<<<<<<
 * 
 * VERSION = ZMQ_VERSION
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(PYZMQ_DRAFT_API); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DRAFT_API, __pyx_t_1) < 0) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":7
 * DRAFT_API = PYZMQ_DRAFT_API
 * 
 * VERSION = ZMQ_VERSION             # <<<<<<<<<<<<<<
 * VERSION_MAJOR = ZMQ_VERSION_MAJOR
 * VERSION_MINOR = ZMQ_VERSION_MINOR
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VERSION); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VERSION, __pyx_t_1) < 0) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":8
 * 
 * VERSION = ZMQ_VERSION
 * VERSION_MAJOR = ZMQ_VERSION_MAJOR             # <<<<<<<<<<<<<<
 * VERSION_MINOR = ZMQ_VERSION_MINOR
 * VERSION_PATCH = ZMQ_VERSION_PATCH
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VERSION_MAJOR); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VERSION_MAJOR, __pyx_t_1) < 0) __PYX_ERR(1, 8, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":9
 * VERSION = ZMQ_VERSION
 * VERSION_MAJOR = ZMQ_VERSION_MAJOR
 * VERSION_MINOR = ZMQ_VERSION_MINOR             # <<<<<<<<<<<<<<
 * VERSION_PATCH = ZMQ_VERSION_PATCH
 * NOBLOCK = ZMQ_NOBLOCK
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VERSION_MINOR); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VERSION_MINOR, __pyx_t_1) < 0) __PYX_ERR(1, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":10
 * VERSION_MAJOR = ZMQ_VERSION_MAJOR
 * VERSION_MINOR = ZMQ_VERSION_MINOR
 * VERSION_PATCH = ZMQ_VERSION_PATCH             # <<<<<<<<<<<<<<
 * NOBLOCK = ZMQ_NOBLOCK
 * DONTWAIT = ZMQ_DONTWAIT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VERSION_PATCH); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VERSION_PATCH, __pyx_t_1) < 0) __PYX_ERR(1, 10, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":11
 * VERSION_MINOR = ZMQ_VERSION_MINOR
 * VERSION_PATCH = ZMQ_VERSION_PATCH
 * NOBLOCK = ZMQ_NOBLOCK             # <<<<<<<<<<<<<<
 * DONTWAIT = ZMQ_DONTWAIT
 * POLLIN = ZMQ_POLLIN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_NOBLOCK); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NOBLOCK, __pyx_t_1) < 0) __PYX_ERR(1, 11, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":12
 * VERSION_PATCH = ZMQ_VERSION_PATCH
 * NOBLOCK = ZMQ_NOBLOCK
 * DONTWAIT = ZMQ_DONTWAIT             # <<<<<<<<<<<<<<
 * POLLIN = ZMQ_POLLIN
 * POLLOUT = ZMQ_POLLOUT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DONTWAIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DONTWAIT, __pyx_t_1) < 0) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":13
 * NOBLOCK = ZMQ_NOBLOCK
 * DONTWAIT = ZMQ_DONTWAIT
 * POLLIN = ZMQ_POLLIN             # <<<<<<<<<<<<<<
 * POLLOUT = ZMQ_POLLOUT
 * POLLERR = ZMQ_POLLERR
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_POLLIN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POLLIN, __pyx_t_1) < 0) __PYX_ERR(1, 13, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":14
 * DONTWAIT = ZMQ_DONTWAIT
 * POLLIN = ZMQ_POLLIN
 * POLLOUT = ZMQ_POLLOUT             # <<<<<<<<<<<<<<
 * POLLERR = ZMQ_POLLERR
 * POLLPRI = ZMQ_POLLPRI
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_POLLOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POLLOUT, __pyx_t_1) < 0) __PYX_ERR(1, 14, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":15
 * POLLIN = ZMQ_POLLIN
 * POLLOUT = ZMQ_POLLOUT
 * POLLERR = ZMQ_POLLERR             # <<<<<<<<<<<<<<
 * POLLPRI = ZMQ_POLLPRI
 * SNDMORE = ZMQ_SNDMORE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_POLLERR); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POLLERR, __pyx_t_1) < 0) __PYX_ERR(1, 15, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":16
 * POLLOUT = ZMQ_POLLOUT
 * POLLERR = ZMQ_POLLERR
 * POLLPRI = ZMQ_POLLPRI             # <<<<<<<<<<<<<<
 * SNDMORE = ZMQ_SNDMORE
 * STREAMER = ZMQ_STREAMER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_POLLPRI); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POLLPRI, __pyx_t_1) < 0) __PYX_ERR(1, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":17
 * POLLERR = ZMQ_POLLERR
 * POLLPRI = ZMQ_POLLPRI
 * SNDMORE = ZMQ_SNDMORE             # <<<<<<<<<<<<<<
 * STREAMER = ZMQ_STREAMER
 * FORWARDER = ZMQ_FORWARDER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SNDMORE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SNDMORE, __pyx_t_1) < 0) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":18
 * POLLPRI = ZMQ_POLLPRI
 * SNDMORE = ZMQ_SNDMORE
 * STREAMER = ZMQ_STREAMER             # <<<<<<<<<<<<<<
 * FORWARDER = ZMQ_FORWARDER
 * QUEUE = ZMQ_QUEUE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_STREAMER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STREAMER, __pyx_t_1) < 0) __PYX_ERR(1, 18, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":19
 * SNDMORE = ZMQ_SNDMORE
 * STREAMER = ZMQ_STREAMER
 * FORWARDER = ZMQ_FORWARDER             # <<<<<<<<<<<<<<
 * QUEUE = ZMQ_QUEUE
 * IO_THREADS_DFLT = ZMQ_IO_THREADS_DFLT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_FORWARDER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FORWARDER, __pyx_t_1) < 0) __PYX_ERR(1, 19, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":20
 * STREAMER = ZMQ_STREAMER
 * FORWARDER = ZMQ_FORWARDER
 * QUEUE = ZMQ_QUEUE             # <<<<<<<<<<<<<<
 * IO_THREADS_DFLT = ZMQ_IO_THREADS_DFLT
 * MAX_SOCKETS_DFLT = ZMQ_MAX_SOCKETS_DFLT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_QUEUE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_QUEUE, __pyx_t_1) < 0) __PYX_ERR(1, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":21
 * FORWARDER = ZMQ_FORWARDER
 * QUEUE = ZMQ_QUEUE
 * IO_THREADS_DFLT = ZMQ_IO_THREADS_DFLT             # <<<<<<<<<<<<<<
 * MAX_SOCKETS_DFLT = ZMQ_MAX_SOCKETS_DFLT
 * POLLITEMS_DFLT = ZMQ_POLLITEMS_DFLT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IO_THREADS_DFLT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IO_THREADS_DFLT, __pyx_t_1) < 0) __PYX_ERR(1, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":22
 * QUEUE = ZMQ_QUEUE
 * IO_THREADS_DFLT = ZMQ_IO_THREADS_DFLT
 * MAX_SOCKETS_DFLT = ZMQ_MAX_SOCKETS_DFLT             # <<<<<<<<<<<<<<
 * POLLITEMS_DFLT = ZMQ_POLLITEMS_DFLT
 * THREAD_PRIORITY_DFLT = ZMQ_THREAD_PRIORITY_DFLT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MAX_SOCKETS_DFLT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_SOCKETS_DFLT, __pyx_t_1) < 0) __PYX_ERR(1, 22, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":23
 * IO_THREADS_DFLT = ZMQ_IO_THREADS_DFLT
 * MAX_SOCKETS_DFLT = ZMQ_MAX_SOCKETS_DFLT
 * POLLITEMS_DFLT = ZMQ_POLLITEMS_DFLT             # <<<<<<<<<<<<<<
 * THREAD_PRIORITY_DFLT = ZMQ_THREAD_PRIORITY_DFLT
 * THREAD_SCHED_POLICY_DFLT = ZMQ_THREAD_SCHED_POLICY_DFLT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_POLLITEMS_DFLT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_POLLITEMS_DFLT, __pyx_t_1) < 0) __PYX_ERR(1, 23, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":24
 * MAX_SOCKETS_DFLT = ZMQ_MAX_SOCKETS_DFLT
 * POLLITEMS_DFLT = ZMQ_POLLITEMS_DFLT
 * THREAD_PRIORITY_DFLT = ZMQ_THREAD_PRIORITY_DFLT             # <<<<<<<<<<<<<<
 * THREAD_SCHED_POLICY_DFLT = ZMQ_THREAD_SCHED_POLICY_DFLT
 * PAIR = ZMQ_PAIR
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_THREAD_PRIORITY_DFLT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THREAD_PRIORITY_DFLT, __pyx_t_1) < 0) __PYX_ERR(1, 24, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":25
 * POLLITEMS_DFLT = ZMQ_POLLITEMS_DFLT
 * THREAD_PRIORITY_DFLT = ZMQ_THREAD_PRIORITY_DFLT
 * THREAD_SCHED_POLICY_DFLT = ZMQ_THREAD_SCHED_POLICY_DFLT             # <<<<<<<<<<<<<<
 * PAIR = ZMQ_PAIR
 * PUB = ZMQ_PUB
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_THREAD_SCHED_POLICY_DFLT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THREAD_SCHED_POLICY_DFLT, __pyx_t_1) < 0) __PYX_ERR(1, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":26
 * THREAD_PRIORITY_DFLT = ZMQ_THREAD_PRIORITY_DFLT
 * THREAD_SCHED_POLICY_DFLT = ZMQ_THREAD_SCHED_POLICY_DFLT
 * PAIR = ZMQ_PAIR             # <<<<<<<<<<<<<<
 * PUB = ZMQ_PUB
 * SUB = ZMQ_SUB
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PAIR); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PAIR, __pyx_t_1) < 0) __PYX_ERR(1, 26, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":27
 * THREAD_SCHED_POLICY_DFLT = ZMQ_THREAD_SCHED_POLICY_DFLT
 * PAIR = ZMQ_PAIR
 * PUB = ZMQ_PUB             # <<<<<<<<<<<<<<
 * SUB = ZMQ_SUB
 * REQ = ZMQ_REQ
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PUB); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PUB, __pyx_t_1) < 0) __PYX_ERR(1, 27, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":28
 * PAIR = ZMQ_PAIR
 * PUB = ZMQ_PUB
 * SUB = ZMQ_SUB             # <<<<<<<<<<<<<<
 * REQ = ZMQ_REQ
 * REP = ZMQ_REP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SUB); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SUB, __pyx_t_1) < 0) __PYX_ERR(1, 28, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":29
 * PUB = ZMQ_PUB
 * SUB = ZMQ_SUB
 * REQ = ZMQ_REQ             # <<<<<<<<<<<<<<
 * REP = ZMQ_REP
 * DEALER = ZMQ_DEALER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_REQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_REQ, __pyx_t_1) < 0) __PYX_ERR(1, 29, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":30
 * SUB = ZMQ_SUB
 * REQ = ZMQ_REQ
 * REP = ZMQ_REP             # <<<<<<<<<<<<<<
 * DEALER = ZMQ_DEALER
 * ROUTER = ZMQ_ROUTER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_REP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_REP, __pyx_t_1) < 0) __PYX_ERR(1, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":31
 * REQ = ZMQ_REQ
 * REP = ZMQ_REP
 * DEALER = ZMQ_DEALER             # <<<<<<<<<<<<<<
 * ROUTER = ZMQ_ROUTER
 * XREQ = ZMQ_XREQ
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DEALER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DEALER, __pyx_t_1) < 0) __PYX_ERR(1, 31, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":32
 * REP = ZMQ_REP
 * DEALER = ZMQ_DEALER
 * ROUTER = ZMQ_ROUTER             # <<<<<<<<<<<<<<
 * XREQ = ZMQ_XREQ
 * XREP = ZMQ_XREP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ROUTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER, __pyx_t_1) < 0) __PYX_ERR(1, 32, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":33
 * DEALER = ZMQ_DEALER
 * ROUTER = ZMQ_ROUTER
 * XREQ = ZMQ_XREQ             # <<<<<<<<<<<<<<
 * XREP = ZMQ_XREP
 * PULL = ZMQ_PULL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XREQ); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XREQ, __pyx_t_1) < 0) __PYX_ERR(1, 33, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":34
 * ROUTER = ZMQ_ROUTER
 * XREQ = ZMQ_XREQ
 * XREP = ZMQ_XREP             # <<<<<<<<<<<<<<
 * PULL = ZMQ_PULL
 * PUSH = ZMQ_PUSH
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XREP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XREP, __pyx_t_1) < 0) __PYX_ERR(1, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":35
 * XREQ = ZMQ_XREQ
 * XREP = ZMQ_XREP
 * PULL = ZMQ_PULL             # <<<<<<<<<<<<<<
 * PUSH = ZMQ_PUSH
 * XPUB = ZMQ_XPUB
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PULL, __pyx_t_1) < 0) __PYX_ERR(1, 35, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":36
 * XREP = ZMQ_XREP
 * PULL = ZMQ_PULL
 * PUSH = ZMQ_PUSH             # <<<<<<<<<<<<<<
 * XPUB = ZMQ_XPUB
 * XSUB = ZMQ_XSUB
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PUSH); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PUSH, __pyx_t_1) < 0) __PYX_ERR(1, 36, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":37
 * PULL = ZMQ_PULL
 * PUSH = ZMQ_PUSH
 * XPUB = ZMQ_XPUB             # <<<<<<<<<<<<<<
 * XSUB = ZMQ_XSUB
 * UPSTREAM = ZMQ_UPSTREAM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB, __pyx_t_1) < 0) __PYX_ERR(1, 37, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":38
 * PUSH = ZMQ_PUSH
 * XPUB = ZMQ_XPUB
 * XSUB = ZMQ_XSUB             # <<<<<<<<<<<<<<
 * UPSTREAM = ZMQ_UPSTREAM
 * DOWNSTREAM = ZMQ_DOWNSTREAM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XSUB); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XSUB, __pyx_t_1) < 0) __PYX_ERR(1, 38, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":39
 * XPUB = ZMQ_XPUB
 * XSUB = ZMQ_XSUB
 * UPSTREAM = ZMQ_UPSTREAM             # <<<<<<<<<<<<<<
 * DOWNSTREAM = ZMQ_DOWNSTREAM
 * STREAM = ZMQ_STREAM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_UPSTREAM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UPSTREAM, __pyx_t_1) < 0) __PYX_ERR(1, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":40
 * XSUB = ZMQ_XSUB
 * UPSTREAM = ZMQ_UPSTREAM
 * DOWNSTREAM = ZMQ_DOWNSTREAM             # <<<<<<<<<<<<<<
 * STREAM = ZMQ_STREAM
 * SERVER = ZMQ_SERVER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DOWNSTREAM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DOWNSTREAM, __pyx_t_1) < 0) __PYX_ERR(1, 40, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":41
 * UPSTREAM = ZMQ_UPSTREAM
 * DOWNSTREAM = ZMQ_DOWNSTREAM
 * STREAM = ZMQ_STREAM             # <<<<<<<<<<<<<<
 * SERVER = ZMQ_SERVER
 * CLIENT = ZMQ_CLIENT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_STREAM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STREAM, __pyx_t_1) < 0) __PYX_ERR(1, 41, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":42
 * DOWNSTREAM = ZMQ_DOWNSTREAM
 * STREAM = ZMQ_STREAM
 * SERVER = ZMQ_SERVER             # <<<<<<<<<<<<<<
 * CLIENT = ZMQ_CLIENT
 * RADIO = ZMQ_RADIO
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SERVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SERVER, __pyx_t_1) < 0) __PYX_ERR(1, 42, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":43
 * STREAM = ZMQ_STREAM
 * SERVER = ZMQ_SERVER
 * CLIENT = ZMQ_CLIENT             # <<<<<<<<<<<<<<
 * RADIO = ZMQ_RADIO
 * DISH = ZMQ_DISH
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CLIENT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CLIENT, __pyx_t_1) < 0) __PYX_ERR(1, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":44
 * SERVER = ZMQ_SERVER
 * CLIENT = ZMQ_CLIENT
 * RADIO = ZMQ_RADIO             # <<<<<<<<<<<<<<
 * DISH = ZMQ_DISH
 * GATHER = ZMQ_GATHER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RADIO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RADIO, __pyx_t_1) < 0) __PYX_ERR(1, 44, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":45
 * CLIENT = ZMQ_CLIENT
 * RADIO = ZMQ_RADIO
 * DISH = ZMQ_DISH             # <<<<<<<<<<<<<<
 * GATHER = ZMQ_GATHER
 * SCATTER = ZMQ_SCATTER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DISH); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DISH, __pyx_t_1) < 0) __PYX_ERR(1, 45, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":46
 * RADIO = ZMQ_RADIO
 * DISH = ZMQ_DISH
 * GATHER = ZMQ_GATHER             # <<<<<<<<<<<<<<
 * SCATTER = ZMQ_SCATTER
 * DGRAM = ZMQ_DGRAM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GATHER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GATHER, __pyx_t_1) < 0) __PYX_ERR(1, 46, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":47
 * DISH = ZMQ_DISH
 * GATHER = ZMQ_GATHER
 * SCATTER = ZMQ_SCATTER             # <<<<<<<<<<<<<<
 * DGRAM = ZMQ_DGRAM
 * EVENT_CONNECTED = ZMQ_EVENT_CONNECTED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SCATTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SCATTER, __pyx_t_1) < 0) __PYX_ERR(1, 47, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":48
 * GATHER = ZMQ_GATHER
 * SCATTER = ZMQ_SCATTER
 * DGRAM = ZMQ_DGRAM             # <<<<<<<<<<<<<<
 * EVENT_CONNECTED = ZMQ_EVENT_CONNECTED
 * EVENT_CONNECT_DELAYED = ZMQ_EVENT_CONNECT_DELAYED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DGRAM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DGRAM, __pyx_t_1) < 0) __PYX_ERR(1, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":49
 * SCATTER = ZMQ_SCATTER
 * DGRAM = ZMQ_DGRAM
 * EVENT_CONNECTED = ZMQ_EVENT_CONNECTED             # <<<<<<<<<<<<<<
 * EVENT_CONNECT_DELAYED = ZMQ_EVENT_CONNECT_DELAYED
 * EVENT_CONNECT_RETRIED = ZMQ_EVENT_CONNECT_RETRIED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_CONNECTED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_CONNECTED, __pyx_t_1) < 0) __PYX_ERR(1, 49, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":50
 * DGRAM = ZMQ_DGRAM
 * EVENT_CONNECTED = ZMQ_EVENT_CONNECTED
 * EVENT_CONNECT_DELAYED = ZMQ_EVENT_CONNECT_DELAYED             # <<<<<<<<<<<<<<
 * EVENT_CONNECT_RETRIED = ZMQ_EVENT_CONNECT_RETRIED
 * EVENT_LISTENING = ZMQ_EVENT_LISTENING
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_CONNECT_DELAYED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_CONNECT_DELAYED, __pyx_t_1) < 0) __PYX_ERR(1, 50, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":51
 * EVENT_CONNECTED = ZMQ_EVENT_CONNECTED
 * EVENT_CONNECT_DELAYED = ZMQ_EVENT_CONNECT_DELAYED
 * EVENT_CONNECT_RETRIED = ZMQ_EVENT_CONNECT_RETRIED             # <<<<<<<<<<<<<<
 * EVENT_LISTENING = ZMQ_EVENT_LISTENING
 * EVENT_BIND_FAILED = ZMQ_EVENT_BIND_FAILED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_CONNECT_RETRIED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_CONNECT_RETRIED, __pyx_t_1) < 0) __PYX_ERR(1, 51, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":52
 * EVENT_CONNECT_DELAYED = ZMQ_EVENT_CONNECT_DELAYED
 * EVENT_CONNECT_RETRIED = ZMQ_EVENT_CONNECT_RETRIED
 * EVENT_LISTENING = ZMQ_EVENT_LISTENING             # <<<<<<<<<<<<<<
 * EVENT_BIND_FAILED = ZMQ_EVENT_BIND_FAILED
 * EVENT_ACCEPTED = ZMQ_EVENT_ACCEPTED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_LISTENING); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_LISTENING, __pyx_t_1) < 0) __PYX_ERR(1, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":53
 * EVENT_CONNECT_RETRIED = ZMQ_EVENT_CONNECT_RETRIED
 * EVENT_LISTENING = ZMQ_EVENT_LISTENING
 * EVENT_BIND_FAILED = ZMQ_EVENT_BIND_FAILED             # <<<<<<<<<<<<<<
 * EVENT_ACCEPTED = ZMQ_EVENT_ACCEPTED
 * EVENT_ACCEPT_FAILED = ZMQ_EVENT_ACCEPT_FAILED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_BIND_FAILED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_BIND_FAILED, __pyx_t_1) < 0) __PYX_ERR(1, 53, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":54
 * EVENT_LISTENING = ZMQ_EVENT_LISTENING
 * EVENT_BIND_FAILED = ZMQ_EVENT_BIND_FAILED
 * EVENT_ACCEPTED = ZMQ_EVENT_ACCEPTED             # <<<<<<<<<<<<<<
 * EVENT_ACCEPT_FAILED = ZMQ_EVENT_ACCEPT_FAILED
 * EVENT_CLOSED = ZMQ_EVENT_CLOSED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_ACCEPTED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_ACCEPTED, __pyx_t_1) < 0) __PYX_ERR(1, 54, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":55
 * EVENT_BIND_FAILED = ZMQ_EVENT_BIND_FAILED
 * EVENT_ACCEPTED = ZMQ_EVENT_ACCEPTED
 * EVENT_ACCEPT_FAILED = ZMQ_EVENT_ACCEPT_FAILED             # <<<<<<<<<<<<<<
 * EVENT_CLOSED = ZMQ_EVENT_CLOSED
 * EVENT_CLOSE_FAILED = ZMQ_EVENT_CLOSE_FAILED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_ACCEPT_FAILED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_ACCEPT_FAILED, __pyx_t_1) < 0) __PYX_ERR(1, 55, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":56
 * EVENT_ACCEPTED = ZMQ_EVENT_ACCEPTED
 * EVENT_ACCEPT_FAILED = ZMQ_EVENT_ACCEPT_FAILED
 * EVENT_CLOSED = ZMQ_EVENT_CLOSED             # <<<<<<<<<<<<<<
 * EVENT_CLOSE_FAILED = ZMQ_EVENT_CLOSE_FAILED
 * EVENT_DISCONNECTED = ZMQ_EVENT_DISCONNECTED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_CLOSED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_CLOSED, __pyx_t_1) < 0) __PYX_ERR(1, 56, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":57
 * EVENT_ACCEPT_FAILED = ZMQ_EVENT_ACCEPT_FAILED
 * EVENT_CLOSED = ZMQ_EVENT_CLOSED
 * EVENT_CLOSE_FAILED = ZMQ_EVENT_CLOSE_FAILED             # <<<<<<<<<<<<<<
 * EVENT_DISCONNECTED = ZMQ_EVENT_DISCONNECTED
 * EVENT_ALL = ZMQ_EVENT_ALL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_CLOSE_FAILED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_CLOSE_FAILED, __pyx_t_1) < 0) __PYX_ERR(1, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":58
 * EVENT_CLOSED = ZMQ_EVENT_CLOSED
 * EVENT_CLOSE_FAILED = ZMQ_EVENT_CLOSE_FAILED
 * EVENT_DISCONNECTED = ZMQ_EVENT_DISCONNECTED             # <<<<<<<<<<<<<<
 * EVENT_ALL = ZMQ_EVENT_ALL
 * EVENT_MONITOR_STOPPED = ZMQ_EVENT_MONITOR_STOPPED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_DISCONNECTED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_DISCONNECTED, __pyx_t_1) < 0) __PYX_ERR(1, 58, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":59
 * EVENT_CLOSE_FAILED = ZMQ_EVENT_CLOSE_FAILED
 * EVENT_DISCONNECTED = ZMQ_EVENT_DISCONNECTED
 * EVENT_ALL = ZMQ_EVENT_ALL             # <<<<<<<<<<<<<<
 * EVENT_MONITOR_STOPPED = ZMQ_EVENT_MONITOR_STOPPED
 * globals()['NULL'] = ZMQ_NULL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_ALL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_ALL, __pyx_t_1) < 0) __PYX_ERR(1, 59, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":60
 * EVENT_DISCONNECTED = ZMQ_EVENT_DISCONNECTED
 * EVENT_ALL = ZMQ_EVENT_ALL
 * EVENT_MONITOR_STOPPED = ZMQ_EVENT_MONITOR_STOPPED             # <<<<<<<<<<<<<<
 * globals()['NULL'] = ZMQ_NULL
 * PLAIN = ZMQ_PLAIN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENT_MONITOR_STOPPED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENT_MONITOR_STOPPED, __pyx_t_1) < 0) __PYX_ERR(1, 60, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":61
 * EVENT_ALL = ZMQ_EVENT_ALL
 * EVENT_MONITOR_STOPPED = ZMQ_EVENT_MONITOR_STOPPED
 * globals()['NULL'] = ZMQ_NULL             # <<<<<<<<<<<<<<
 * PLAIN = ZMQ_PLAIN
 * CURVE = ZMQ_CURVE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_Globals(); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (unlikely(PyObject_SetItem(__pyx_t_2, __pyx_n_s_NULL, __pyx_t_1) < 0)) __PYX_ERR(1, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":62
 * EVENT_MONITOR_STOPPED = ZMQ_EVENT_MONITOR_STOPPED
 * globals()['NULL'] = ZMQ_NULL
 * PLAIN = ZMQ_PLAIN             # <<<<<<<<<<<<<<
 * CURVE = ZMQ_CURVE
 * GSSAPI = ZMQ_GSSAPI
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PLAIN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PLAIN, __pyx_t_1) < 0) __PYX_ERR(1, 62, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":63
 * globals()['NULL'] = ZMQ_NULL
 * PLAIN = ZMQ_PLAIN
 * CURVE = ZMQ_CURVE             # <<<<<<<<<<<<<<
 * GSSAPI = ZMQ_GSSAPI
 * EAGAIN = ZMQ_EAGAIN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CURVE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CURVE, __pyx_t_1) < 0) __PYX_ERR(1, 63, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":64
 * PLAIN = ZMQ_PLAIN
 * CURVE = ZMQ_CURVE
 * GSSAPI = ZMQ_GSSAPI             # <<<<<<<<<<<<<<
 * EAGAIN = ZMQ_EAGAIN
 * EINVAL = ZMQ_EINVAL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GSSAPI); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GSSAPI, __pyx_t_1) < 0) __PYX_ERR(1, 64, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":65
 * CURVE = ZMQ_CURVE
 * GSSAPI = ZMQ_GSSAPI
 * EAGAIN = ZMQ_EAGAIN             # <<<<<<<<<<<<<<
 * EINVAL = ZMQ_EINVAL
 * EFAULT = ZMQ_EFAULT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EAGAIN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EAGAIN, __pyx_t_1) < 0) __PYX_ERR(1, 65, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":66
 * GSSAPI = ZMQ_GSSAPI
 * EAGAIN = ZMQ_EAGAIN
 * EINVAL = ZMQ_EINVAL             # <<<<<<<<<<<<<<
 * EFAULT = ZMQ_EFAULT
 * ENOMEM = ZMQ_ENOMEM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EINVAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EINVAL, __pyx_t_1) < 0) __PYX_ERR(1, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":67
 * EAGAIN = ZMQ_EAGAIN
 * EINVAL = ZMQ_EINVAL
 * EFAULT = ZMQ_EFAULT             # <<<<<<<<<<<<<<
 * ENOMEM = ZMQ_ENOMEM
 * ENODEV = ZMQ_ENODEV
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EFAULT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EFAULT, __pyx_t_1) < 0) __PYX_ERR(1, 67, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":68
 * EINVAL = ZMQ_EINVAL
 * EFAULT = ZMQ_EFAULT
 * ENOMEM = ZMQ_ENOMEM             # <<<<<<<<<<<<<<
 * ENODEV = ZMQ_ENODEV
 * EMSGSIZE = ZMQ_EMSGSIZE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOMEM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOMEM, __pyx_t_1) < 0) __PYX_ERR(1, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":69
 * EFAULT = ZMQ_EFAULT
 * ENOMEM = ZMQ_ENOMEM
 * ENODEV = ZMQ_ENODEV             # <<<<<<<<<<<<<<
 * EMSGSIZE = ZMQ_EMSGSIZE
 * EAFNOSUPPORT = ZMQ_EAFNOSUPPORT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENODEV); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENODEV, __pyx_t_1) < 0) __PYX_ERR(1, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":70
 * ENOMEM = ZMQ_ENOMEM
 * ENODEV = ZMQ_ENODEV
 * EMSGSIZE = ZMQ_EMSGSIZE             # <<<<<<<<<<<<<<
 * EAFNOSUPPORT = ZMQ_EAFNOSUPPORT
 * ENETUNREACH = ZMQ_ENETUNREACH
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EMSGSIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EMSGSIZE, __pyx_t_1) < 0) __PYX_ERR(1, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":71
 * ENODEV = ZMQ_ENODEV
 * EMSGSIZE = ZMQ_EMSGSIZE
 * EAFNOSUPPORT = ZMQ_EAFNOSUPPORT             # <<<<<<<<<<<<<<
 * ENETUNREACH = ZMQ_ENETUNREACH
 * ECONNABORTED = ZMQ_ECONNABORTED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EAFNOSUPPORT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EAFNOSUPPORT, __pyx_t_1) < 0) __PYX_ERR(1, 71, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":72
 * EMSGSIZE = ZMQ_EMSGSIZE
 * EAFNOSUPPORT = ZMQ_EAFNOSUPPORT
 * ENETUNREACH = ZMQ_ENETUNREACH             # <<<<<<<<<<<<<<
 * ECONNABORTED = ZMQ_ECONNABORTED
 * ECONNRESET = ZMQ_ECONNRESET
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENETUNREACH); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENETUNREACH, __pyx_t_1) < 0) __PYX_ERR(1, 72, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":73
 * EAFNOSUPPORT = ZMQ_EAFNOSUPPORT
 * ENETUNREACH = ZMQ_ENETUNREACH
 * ECONNABORTED = ZMQ_ECONNABORTED             # <<<<<<<<<<<<<<
 * ECONNRESET = ZMQ_ECONNRESET
 * ENOTCONN = ZMQ_ENOTCONN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ECONNABORTED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ECONNABORTED, __pyx_t_1) < 0) __PYX_ERR(1, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":74
 * ENETUNREACH = ZMQ_ENETUNREACH
 * ECONNABORTED = ZMQ_ECONNABORTED
 * ECONNRESET = ZMQ_ECONNRESET             # <<<<<<<<<<<<<<
 * ENOTCONN = ZMQ_ENOTCONN
 * ETIMEDOUT = ZMQ_ETIMEDOUT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ECONNRESET); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ECONNRESET, __pyx_t_1) < 0) __PYX_ERR(1, 74, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":75
 * ECONNABORTED = ZMQ_ECONNABORTED
 * ECONNRESET = ZMQ_ECONNRESET
 * ENOTCONN = ZMQ_ENOTCONN             # <<<<<<<<<<<<<<
 * ETIMEDOUT = ZMQ_ETIMEDOUT
 * EHOSTUNREACH = ZMQ_EHOSTUNREACH
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOTCONN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOTCONN, __pyx_t_1) < 0) __PYX_ERR(1, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":76
 * ECONNRESET = ZMQ_ECONNRESET
 * ENOTCONN = ZMQ_ENOTCONN
 * ETIMEDOUT = ZMQ_ETIMEDOUT             # <<<<<<<<<<<<<<
 * EHOSTUNREACH = ZMQ_EHOSTUNREACH
 * ENETRESET = ZMQ_ENETRESET
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ETIMEDOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ETIMEDOUT, __pyx_t_1) < 0) __PYX_ERR(1, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":77
 * ENOTCONN = ZMQ_ENOTCONN
 * ETIMEDOUT = ZMQ_ETIMEDOUT
 * EHOSTUNREACH = ZMQ_EHOSTUNREACH             # <<<<<<<<<<<<<<
 * ENETRESET = ZMQ_ENETRESET
 * HAUSNUMERO = ZMQ_HAUSNUMERO
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EHOSTUNREACH); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EHOSTUNREACH, __pyx_t_1) < 0) __PYX_ERR(1, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":78
 * ETIMEDOUT = ZMQ_ETIMEDOUT
 * EHOSTUNREACH = ZMQ_EHOSTUNREACH
 * ENETRESET = ZMQ_ENETRESET             # <<<<<<<<<<<<<<
 * HAUSNUMERO = ZMQ_HAUSNUMERO
 * ENOTSUP = ZMQ_ENOTSUP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENETRESET); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENETRESET, __pyx_t_1) < 0) __PYX_ERR(1, 78, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":79
 * EHOSTUNREACH = ZMQ_EHOSTUNREACH
 * ENETRESET = ZMQ_ENETRESET
 * HAUSNUMERO = ZMQ_HAUSNUMERO             # <<<<<<<<<<<<<<
 * ENOTSUP = ZMQ_ENOTSUP
 * EPROTONOSUPPORT = ZMQ_EPROTONOSUPPORT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HAUSNUMERO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HAUSNUMERO, __pyx_t_1) < 0) __PYX_ERR(1, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":80
 * ENETRESET = ZMQ_ENETRESET
 * HAUSNUMERO = ZMQ_HAUSNUMERO
 * ENOTSUP = ZMQ_ENOTSUP             # <<<<<<<<<<<<<<
 * EPROTONOSUPPORT = ZMQ_EPROTONOSUPPORT
 * ENOBUFS = ZMQ_ENOBUFS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOTSUP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOTSUP, __pyx_t_1) < 0) __PYX_ERR(1, 80, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":81
 * HAUSNUMERO = ZMQ_HAUSNUMERO
 * ENOTSUP = ZMQ_ENOTSUP
 * EPROTONOSUPPORT = ZMQ_EPROTONOSUPPORT             # <<<<<<<<<<<<<<
 * ENOBUFS = ZMQ_ENOBUFS
 * ENETDOWN = ZMQ_ENETDOWN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EPROTONOSUPPORT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EPROTONOSUPPORT, __pyx_t_1) < 0) __PYX_ERR(1, 81, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":82
 * ENOTSUP = ZMQ_ENOTSUP
 * EPROTONOSUPPORT = ZMQ_EPROTONOSUPPORT
 * ENOBUFS = ZMQ_ENOBUFS             # <<<<<<<<<<<<<<
 * ENETDOWN = ZMQ_ENETDOWN
 * EADDRINUSE = ZMQ_EADDRINUSE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOBUFS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOBUFS, __pyx_t_1) < 0) __PYX_ERR(1, 82, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":83
 * EPROTONOSUPPORT = ZMQ_EPROTONOSUPPORT
 * ENOBUFS = ZMQ_ENOBUFS
 * ENETDOWN = ZMQ_ENETDOWN             # <<<<<<<<<<<<<<
 * EADDRINUSE = ZMQ_EADDRINUSE
 * EADDRNOTAVAIL = ZMQ_EADDRNOTAVAIL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENETDOWN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENETDOWN, __pyx_t_1) < 0) __PYX_ERR(1, 83, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":84
 * ENOBUFS = ZMQ_ENOBUFS
 * ENETDOWN = ZMQ_ENETDOWN
 * EADDRINUSE = ZMQ_EADDRINUSE             # <<<<<<<<<<<<<<
 * EADDRNOTAVAIL = ZMQ_EADDRNOTAVAIL
 * ECONNREFUSED = ZMQ_ECONNREFUSED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EADDRINUSE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EADDRINUSE, __pyx_t_1) < 0) __PYX_ERR(1, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":85
 * ENETDOWN = ZMQ_ENETDOWN
 * EADDRINUSE = ZMQ_EADDRINUSE
 * EADDRNOTAVAIL = ZMQ_EADDRNOTAVAIL             # <<<<<<<<<<<<<<
 * ECONNREFUSED = ZMQ_ECONNREFUSED
 * EINPROGRESS = ZMQ_EINPROGRESS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EADDRNOTAVAIL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EADDRNOTAVAIL, __pyx_t_1) < 0) __PYX_ERR(1, 85, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":86
 * EADDRINUSE = ZMQ_EADDRINUSE
 * EADDRNOTAVAIL = ZMQ_EADDRNOTAVAIL
 * ECONNREFUSED = ZMQ_ECONNREFUSED             # <<<<<<<<<<<<<<
 * EINPROGRESS = ZMQ_EINPROGRESS
 * ENOTSOCK = ZMQ_ENOTSOCK
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ECONNREFUSED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ECONNREFUSED, __pyx_t_1) < 0) __PYX_ERR(1, 86, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":87
 * EADDRNOTAVAIL = ZMQ_EADDRNOTAVAIL
 * ECONNREFUSED = ZMQ_ECONNREFUSED
 * EINPROGRESS = ZMQ_EINPROGRESS             # <<<<<<<<<<<<<<
 * ENOTSOCK = ZMQ_ENOTSOCK
 * EFSM = ZMQ_EFSM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EINPROGRESS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EINPROGRESS, __pyx_t_1) < 0) __PYX_ERR(1, 87, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":88
 * ECONNREFUSED = ZMQ_ECONNREFUSED
 * EINPROGRESS = ZMQ_EINPROGRESS
 * ENOTSOCK = ZMQ_ENOTSOCK             # <<<<<<<<<<<<<<
 * EFSM = ZMQ_EFSM
 * ENOCOMPATPROTO = ZMQ_ENOCOMPATPROTO
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOTSOCK); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOTSOCK, __pyx_t_1) < 0) __PYX_ERR(1, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":89
 * EINPROGRESS = ZMQ_EINPROGRESS
 * ENOTSOCK = ZMQ_ENOTSOCK
 * EFSM = ZMQ_EFSM             # <<<<<<<<<<<<<<
 * ENOCOMPATPROTO = ZMQ_ENOCOMPATPROTO
 * ETERM = ZMQ_ETERM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EFSM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EFSM, __pyx_t_1) < 0) __PYX_ERR(1, 89, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":90
 * ENOTSOCK = ZMQ_ENOTSOCK
 * EFSM = ZMQ_EFSM
 * ENOCOMPATPROTO = ZMQ_ENOCOMPATPROTO             # <<<<<<<<<<<<<<
 * ETERM = ZMQ_ETERM
 * EMTHREAD = ZMQ_EMTHREAD
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ENOCOMPATPROTO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ENOCOMPATPROTO, __pyx_t_1) < 0) __PYX_ERR(1, 90, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":91
 * EFSM = ZMQ_EFSM
 * ENOCOMPATPROTO = ZMQ_ENOCOMPATPROTO
 * ETERM = ZMQ_ETERM             # <<<<<<<<<<<<<<
 * EMTHREAD = ZMQ_EMTHREAD
 * IO_THREADS = ZMQ_IO_THREADS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ETERM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ETERM, __pyx_t_1) < 0) __PYX_ERR(1, 91, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":92
 * ENOCOMPATPROTO = ZMQ_ENOCOMPATPROTO
 * ETERM = ZMQ_ETERM
 * EMTHREAD = ZMQ_EMTHREAD             # <<<<<<<<<<<<<<
 * IO_THREADS = ZMQ_IO_THREADS
 * MAX_SOCKETS = ZMQ_MAX_SOCKETS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(EMTHREAD); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EMTHREAD, __pyx_t_1) < 0) __PYX_ERR(1, 92, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":93
 * ETERM = ZMQ_ETERM
 * EMTHREAD = ZMQ_EMTHREAD
 * IO_THREADS = ZMQ_IO_THREADS             # <<<<<<<<<<<<<<
 * MAX_SOCKETS = ZMQ_MAX_SOCKETS
 * SOCKET_LIMIT = ZMQ_SOCKET_LIMIT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IO_THREADS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IO_THREADS, __pyx_t_1) < 0) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":94
 * EMTHREAD = ZMQ_EMTHREAD
 * IO_THREADS = ZMQ_IO_THREADS
 * MAX_SOCKETS = ZMQ_MAX_SOCKETS             # <<<<<<<<<<<<<<
 * SOCKET_LIMIT = ZMQ_SOCKET_LIMIT
 * THREAD_PRIORITY = ZMQ_THREAD_PRIORITY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MAX_SOCKETS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAX_SOCKETS, __pyx_t_1) < 0) __PYX_ERR(1, 94, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":95
 * IO_THREADS = ZMQ_IO_THREADS
 * MAX_SOCKETS = ZMQ_MAX_SOCKETS
 * SOCKET_LIMIT = ZMQ_SOCKET_LIMIT             # <<<<<<<<<<<<<<
 * THREAD_PRIORITY = ZMQ_THREAD_PRIORITY
 * THREAD_SCHED_POLICY = ZMQ_THREAD_SCHED_POLICY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SOCKET_LIMIT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SOCKET_LIMIT, __pyx_t_1) < 0) __PYX_ERR(1, 95, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":96
 * MAX_SOCKETS = ZMQ_MAX_SOCKETS
 * SOCKET_LIMIT = ZMQ_SOCKET_LIMIT
 * THREAD_PRIORITY = ZMQ_THREAD_PRIORITY             # <<<<<<<<<<<<<<
 * THREAD_SCHED_POLICY = ZMQ_THREAD_SCHED_POLICY
 * BLOCKY = ZMQ_BLOCKY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_THREAD_PRIORITY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THREAD_PRIORITY, __pyx_t_1) < 0) __PYX_ERR(1, 96, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":97
 * SOCKET_LIMIT = ZMQ_SOCKET_LIMIT
 * THREAD_PRIORITY = ZMQ_THREAD_PRIORITY
 * THREAD_SCHED_POLICY = ZMQ_THREAD_SCHED_POLICY             # <<<<<<<<<<<<<<
 * BLOCKY = ZMQ_BLOCKY
 * IDENTITY = ZMQ_IDENTITY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_THREAD_SCHED_POLICY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THREAD_SCHED_POLICY, __pyx_t_1) < 0) __PYX_ERR(1, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":98
 * THREAD_PRIORITY = ZMQ_THREAD_PRIORITY
 * THREAD_SCHED_POLICY = ZMQ_THREAD_SCHED_POLICY
 * BLOCKY = ZMQ_BLOCKY             # <<<<<<<<<<<<<<
 * IDENTITY = ZMQ_IDENTITY
 * SUBSCRIBE = ZMQ_SUBSCRIBE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_BLOCKY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BLOCKY, __pyx_t_1) < 0) __PYX_ERR(1, 98, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":99
 * THREAD_SCHED_POLICY = ZMQ_THREAD_SCHED_POLICY
 * BLOCKY = ZMQ_BLOCKY
 * IDENTITY = ZMQ_IDENTITY             # <<<<<<<<<<<<<<
 * SUBSCRIBE = ZMQ_SUBSCRIBE
 * UNSUBSCRIBE = ZMQ_UNSUBSCRIBE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IDENTITY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IDENTITY, __pyx_t_1) < 0) __PYX_ERR(1, 99, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":100
 * BLOCKY = ZMQ_BLOCKY
 * IDENTITY = ZMQ_IDENTITY
 * SUBSCRIBE = ZMQ_SUBSCRIBE             # <<<<<<<<<<<<<<
 * UNSUBSCRIBE = ZMQ_UNSUBSCRIBE
 * LAST_ENDPOINT = ZMQ_LAST_ENDPOINT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SUBSCRIBE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SUBSCRIBE, __pyx_t_1) < 0) __PYX_ERR(1, 100, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":101
 * IDENTITY = ZMQ_IDENTITY
 * SUBSCRIBE = ZMQ_SUBSCRIBE
 * UNSUBSCRIBE = ZMQ_UNSUBSCRIBE             # <<<<<<<<<<<<<<
 * LAST_ENDPOINT = ZMQ_LAST_ENDPOINT
 * TCP_ACCEPT_FILTER = ZMQ_TCP_ACCEPT_FILTER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_UNSUBSCRIBE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_UNSUBSCRIBE, __pyx_t_1) < 0) __PYX_ERR(1, 101, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":102
 * SUBSCRIBE = ZMQ_SUBSCRIBE
 * UNSUBSCRIBE = ZMQ_UNSUBSCRIBE
 * LAST_ENDPOINT = ZMQ_LAST_ENDPOINT             # <<<<<<<<<<<<<<
 * TCP_ACCEPT_FILTER = ZMQ_TCP_ACCEPT_FILTER
 * PLAIN_USERNAME = ZMQ_PLAIN_USERNAME
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_LAST_ENDPOINT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LAST_ENDPOINT, __pyx_t_1) < 0) __PYX_ERR(1, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":103
 * UNSUBSCRIBE = ZMQ_UNSUBSCRIBE
 * LAST_ENDPOINT = ZMQ_LAST_ENDPOINT
 * TCP_ACCEPT_FILTER = ZMQ_TCP_ACCEPT_FILTER             # <<<<<<<<<<<<<<
 * PLAIN_USERNAME = ZMQ_PLAIN_USERNAME
 * PLAIN_PASSWORD = ZMQ_PLAIN_PASSWORD
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_ACCEPT_FILTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_ACCEPT_FILTER, __pyx_t_1) < 0) __PYX_ERR(1, 103, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":104
 * LAST_ENDPOINT = ZMQ_LAST_ENDPOINT
 * TCP_ACCEPT_FILTER = ZMQ_TCP_ACCEPT_FILTER
 * PLAIN_USERNAME = ZMQ_PLAIN_USERNAME             # <<<<<<<<<<<<<<
 * PLAIN_PASSWORD = ZMQ_PLAIN_PASSWORD
 * CURVE_PUBLICKEY = ZMQ_CURVE_PUBLICKEY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PLAIN_USERNAME); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 104, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PLAIN_USERNAME, __pyx_t_1) < 0) __PYX_ERR(1, 104, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":105
 * TCP_ACCEPT_FILTER = ZMQ_TCP_ACCEPT_FILTER
 * PLAIN_USERNAME = ZMQ_PLAIN_USERNAME
 * PLAIN_PASSWORD = ZMQ_PLAIN_PASSWORD             # <<<<<<<<<<<<<<
 * CURVE_PUBLICKEY = ZMQ_CURVE_PUBLICKEY
 * CURVE_SECRETKEY = ZMQ_CURVE_SECRETKEY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PLAIN_PASSWORD); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PLAIN_PASSWORD, __pyx_t_1) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":106
 * PLAIN_USERNAME = ZMQ_PLAIN_USERNAME
 * PLAIN_PASSWORD = ZMQ_PLAIN_PASSWORD
 * CURVE_PUBLICKEY = ZMQ_CURVE_PUBLICKEY             # <<<<<<<<<<<<<<
 * CURVE_SECRETKEY = ZMQ_CURVE_SECRETKEY
 * CURVE_SERVERKEY = ZMQ_CURVE_SERVERKEY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CURVE_PUBLICKEY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CURVE_PUBLICKEY, __pyx_t_1) < 0) __PYX_ERR(1, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":107
 * PLAIN_PASSWORD = ZMQ_PLAIN_PASSWORD
 * CURVE_PUBLICKEY = ZMQ_CURVE_PUBLICKEY
 * CURVE_SECRETKEY = ZMQ_CURVE_SECRETKEY             # <<<<<<<<<<<<<<
 * CURVE_SERVERKEY = ZMQ_CURVE_SERVERKEY
 * ZAP_DOMAIN = ZMQ_ZAP_DOMAIN
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CURVE_SECRETKEY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CURVE_SECRETKEY, __pyx_t_1) < 0) __PYX_ERR(1, 107, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":108
 * CURVE_PUBLICKEY = ZMQ_CURVE_PUBLICKEY
 * CURVE_SECRETKEY = ZMQ_CURVE_SECRETKEY
 * CURVE_SERVERKEY = ZMQ_CURVE_SERVERKEY             # <<<<<<<<<<<<<<
 * ZAP_DOMAIN = ZMQ_ZAP_DOMAIN
 * CONNECT_RID = ZMQ_CONNECT_RID
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CURVE_SERVERKEY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CURVE_SERVERKEY, __pyx_t_1) < 0) __PYX_ERR(1, 108, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":109
 * CURVE_SECRETKEY = ZMQ_CURVE_SECRETKEY
 * CURVE_SERVERKEY = ZMQ_CURVE_SERVERKEY
 * ZAP_DOMAIN = ZMQ_ZAP_DOMAIN             # <<<<<<<<<<<<<<
 * CONNECT_RID = ZMQ_CONNECT_RID
 * GSSAPI_PRINCIPAL = ZMQ_GSSAPI_PRINCIPAL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ZAP_DOMAIN); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ZAP_DOMAIN, __pyx_t_1) < 0) __PYX_ERR(1, 109, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":110
 * CURVE_SERVERKEY = ZMQ_CURVE_SERVERKEY
 * ZAP_DOMAIN = ZMQ_ZAP_DOMAIN
 * CONNECT_RID = ZMQ_CONNECT_RID             # <<<<<<<<<<<<<<
 * GSSAPI_PRINCIPAL = ZMQ_GSSAPI_PRINCIPAL
 * GSSAPI_SERVICE_PRINCIPAL = ZMQ_GSSAPI_SERVICE_PRINCIPAL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CONNECT_RID); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CONNECT_RID, __pyx_t_1) < 0) __PYX_ERR(1, 110, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":111
 * ZAP_DOMAIN = ZMQ_ZAP_DOMAIN
 * CONNECT_RID = ZMQ_CONNECT_RID
 * GSSAPI_PRINCIPAL = ZMQ_GSSAPI_PRINCIPAL             # <<<<<<<<<<<<<<
 * GSSAPI_SERVICE_PRINCIPAL = ZMQ_GSSAPI_SERVICE_PRINCIPAL
 * SOCKS_PROXY = ZMQ_SOCKS_PROXY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GSSAPI_PRINCIPAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GSSAPI_PRINCIPAL, __pyx_t_1) < 0) __PYX_ERR(1, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":112
 * CONNECT_RID = ZMQ_CONNECT_RID
 * GSSAPI_PRINCIPAL = ZMQ_GSSAPI_PRINCIPAL
 * GSSAPI_SERVICE_PRINCIPAL = ZMQ_GSSAPI_SERVICE_PRINCIPAL             # <<<<<<<<<<<<<<
 * SOCKS_PROXY = ZMQ_SOCKS_PROXY
 * XPUB_WELCOME_MSG = ZMQ_XPUB_WELCOME_MSG
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GSSAPI_SERVICE_PRINCIPAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GSSAPI_SERVICE_PRINCIPAL, __pyx_t_1) < 0) __PYX_ERR(1, 112, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":113
 * GSSAPI_PRINCIPAL = ZMQ_GSSAPI_PRINCIPAL
 * GSSAPI_SERVICE_PRINCIPAL = ZMQ_GSSAPI_SERVICE_PRINCIPAL
 * SOCKS_PROXY = ZMQ_SOCKS_PROXY             # <<<<<<<<<<<<<<
 * XPUB_WELCOME_MSG = ZMQ_XPUB_WELCOME_MSG
 * FD = ZMQ_FD
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SOCKS_PROXY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SOCKS_PROXY, __pyx_t_1) < 0) __PYX_ERR(1, 113, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":114
 * GSSAPI_SERVICE_PRINCIPAL = ZMQ_GSSAPI_SERVICE_PRINCIPAL
 * SOCKS_PROXY = ZMQ_SOCKS_PROXY
 * XPUB_WELCOME_MSG = ZMQ_XPUB_WELCOME_MSG             # <<<<<<<<<<<<<<
 * FD = ZMQ_FD
 * RECONNECT_IVL_MAX = ZMQ_RECONNECT_IVL_MAX
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB_WELCOME_MSG); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB_WELCOME_MSG, __pyx_t_1) < 0) __PYX_ERR(1, 114, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":115
 * SOCKS_PROXY = ZMQ_SOCKS_PROXY
 * XPUB_WELCOME_MSG = ZMQ_XPUB_WELCOME_MSG
 * FD = ZMQ_FD             # <<<<<<<<<<<<<<
 * RECONNECT_IVL_MAX = ZMQ_RECONNECT_IVL_MAX
 * SNDTIMEO = ZMQ_SNDTIMEO
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_FD); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FD, __pyx_t_1) < 0) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":116
 * XPUB_WELCOME_MSG = ZMQ_XPUB_WELCOME_MSG
 * FD = ZMQ_FD
 * RECONNECT_IVL_MAX = ZMQ_RECONNECT_IVL_MAX             # <<<<<<<<<<<<<<
 * SNDTIMEO = ZMQ_SNDTIMEO
 * RCVTIMEO = ZMQ_RCVTIMEO
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RECONNECT_IVL_MAX); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECONNECT_IVL_MAX, __pyx_t_1) < 0) __PYX_ERR(1, 116, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":117
 * FD = ZMQ_FD
 * RECONNECT_IVL_MAX = ZMQ_RECONNECT_IVL_MAX
 * SNDTIMEO = ZMQ_SNDTIMEO             # <<<<<<<<<<<<<<
 * RCVTIMEO = ZMQ_RCVTIMEO
 * SNDHWM = ZMQ_SNDHWM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SNDTIMEO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SNDTIMEO, __pyx_t_1) < 0) __PYX_ERR(1, 117, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":118
 * RECONNECT_IVL_MAX = ZMQ_RECONNECT_IVL_MAX
 * SNDTIMEO = ZMQ_SNDTIMEO
 * RCVTIMEO = ZMQ_RCVTIMEO             # <<<<<<<<<<<<<<
 * SNDHWM = ZMQ_SNDHWM
 * RCVHWM = ZMQ_RCVHWM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RCVTIMEO); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RCVTIMEO, __pyx_t_1) < 0) __PYX_ERR(1, 118, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":119
 * SNDTIMEO = ZMQ_SNDTIMEO
 * RCVTIMEO = ZMQ_RCVTIMEO
 * SNDHWM = ZMQ_SNDHWM             # <<<<<<<<<<<<<<
 * RCVHWM = ZMQ_RCVHWM
 * MULTICAST_HOPS = ZMQ_MULTICAST_HOPS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SNDHWM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SNDHWM, __pyx_t_1) < 0) __PYX_ERR(1, 119, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":120
 * RCVTIMEO = ZMQ_RCVTIMEO
 * SNDHWM = ZMQ_SNDHWM
 * RCVHWM = ZMQ_RCVHWM             # <<<<<<<<<<<<<<
 * MULTICAST_HOPS = ZMQ_MULTICAST_HOPS
 * IPV4ONLY = ZMQ_IPV4ONLY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RCVHWM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RCVHWM, __pyx_t_1) < 0) __PYX_ERR(1, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":121
 * SNDHWM = ZMQ_SNDHWM
 * RCVHWM = ZMQ_RCVHWM
 * MULTICAST_HOPS = ZMQ_MULTICAST_HOPS             # <<<<<<<<<<<<<<
 * IPV4ONLY = ZMQ_IPV4ONLY
 * ROUTER_BEHAVIOR = ZMQ_ROUTER_BEHAVIOR
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MULTICAST_HOPS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MULTICAST_HOPS, __pyx_t_1) < 0) __PYX_ERR(1, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":122
 * RCVHWM = ZMQ_RCVHWM
 * MULTICAST_HOPS = ZMQ_MULTICAST_HOPS
 * IPV4ONLY = ZMQ_IPV4ONLY             # <<<<<<<<<<<<<<
 * ROUTER_BEHAVIOR = ZMQ_ROUTER_BEHAVIOR
 * TCP_KEEPALIVE = ZMQ_TCP_KEEPALIVE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IPV4ONLY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IPV4ONLY, __pyx_t_1) < 0) __PYX_ERR(1, 122, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":123
 * MULTICAST_HOPS = ZMQ_MULTICAST_HOPS
 * IPV4ONLY = ZMQ_IPV4ONLY
 * ROUTER_BEHAVIOR = ZMQ_ROUTER_BEHAVIOR             # <<<<<<<<<<<<<<
 * TCP_KEEPALIVE = ZMQ_TCP_KEEPALIVE
 * TCP_KEEPALIVE_CNT = ZMQ_TCP_KEEPALIVE_CNT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ROUTER_BEHAVIOR); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER_BEHAVIOR, __pyx_t_1) < 0) __PYX_ERR(1, 123, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":124
 * IPV4ONLY = ZMQ_IPV4ONLY
 * ROUTER_BEHAVIOR = ZMQ_ROUTER_BEHAVIOR
 * TCP_KEEPALIVE = ZMQ_TCP_KEEPALIVE             # <<<<<<<<<<<<<<
 * TCP_KEEPALIVE_CNT = ZMQ_TCP_KEEPALIVE_CNT
 * TCP_KEEPALIVE_IDLE = ZMQ_TCP_KEEPALIVE_IDLE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_KEEPALIVE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_KEEPALIVE, __pyx_t_1) < 0) __PYX_ERR(1, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":125
 * ROUTER_BEHAVIOR = ZMQ_ROUTER_BEHAVIOR
 * TCP_KEEPALIVE = ZMQ_TCP_KEEPALIVE
 * TCP_KEEPALIVE_CNT = ZMQ_TCP_KEEPALIVE_CNT             # <<<<<<<<<<<<<<
 * TCP_KEEPALIVE_IDLE = ZMQ_TCP_KEEPALIVE_IDLE
 * TCP_KEEPALIVE_INTVL = ZMQ_TCP_KEEPALIVE_INTVL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_KEEPALIVE_CNT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_KEEPALIVE_CNT, __pyx_t_1) < 0) __PYX_ERR(1, 125, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":126
 * TCP_KEEPALIVE = ZMQ_TCP_KEEPALIVE
 * TCP_KEEPALIVE_CNT = ZMQ_TCP_KEEPALIVE_CNT
 * TCP_KEEPALIVE_IDLE = ZMQ_TCP_KEEPALIVE_IDLE             # <<<<<<<<<<<<<<
 * TCP_KEEPALIVE_INTVL = ZMQ_TCP_KEEPALIVE_INTVL
 * DELAY_ATTACH_ON_CONNECT = ZMQ_DELAY_ATTACH_ON_CONNECT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_KEEPALIVE_IDLE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_KEEPALIVE_IDLE, __pyx_t_1) < 0) __PYX_ERR(1, 126, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":127
 * TCP_KEEPALIVE_CNT = ZMQ_TCP_KEEPALIVE_CNT
 * TCP_KEEPALIVE_IDLE = ZMQ_TCP_KEEPALIVE_IDLE
 * TCP_KEEPALIVE_INTVL = ZMQ_TCP_KEEPALIVE_INTVL             # <<<<<<<<<<<<<<
 * DELAY_ATTACH_ON_CONNECT = ZMQ_DELAY_ATTACH_ON_CONNECT
 * XPUB_VERBOSE = ZMQ_XPUB_VERBOSE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_KEEPALIVE_INTVL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_KEEPALIVE_INTVL, __pyx_t_1) < 0) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":128
 * TCP_KEEPALIVE_IDLE = ZMQ_TCP_KEEPALIVE_IDLE
 * TCP_KEEPALIVE_INTVL = ZMQ_TCP_KEEPALIVE_INTVL
 * DELAY_ATTACH_ON_CONNECT = ZMQ_DELAY_ATTACH_ON_CONNECT             # <<<<<<<<<<<<<<
 * XPUB_VERBOSE = ZMQ_XPUB_VERBOSE
 * EVENTS = ZMQ_EVENTS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_DELAY_ATTACH_ON_CONNECT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_DELAY_ATTACH_ON_CONNECT, __pyx_t_1) < 0) __PYX_ERR(1, 128, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":129
 * TCP_KEEPALIVE_INTVL = ZMQ_TCP_KEEPALIVE_INTVL
 * DELAY_ATTACH_ON_CONNECT = ZMQ_DELAY_ATTACH_ON_CONNECT
 * XPUB_VERBOSE = ZMQ_XPUB_VERBOSE             # <<<<<<<<<<<<<<
 * EVENTS = ZMQ_EVENTS
 * TYPE = ZMQ_TYPE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB_VERBOSE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB_VERBOSE, __pyx_t_1) < 0) __PYX_ERR(1, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":130
 * DELAY_ATTACH_ON_CONNECT = ZMQ_DELAY_ATTACH_ON_CONNECT
 * XPUB_VERBOSE = ZMQ_XPUB_VERBOSE
 * EVENTS = ZMQ_EVENTS             # <<<<<<<<<<<<<<
 * TYPE = ZMQ_TYPE
 * LINGER = ZMQ_LINGER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_EVENTS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_EVENTS, __pyx_t_1) < 0) __PYX_ERR(1, 130, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":131
 * XPUB_VERBOSE = ZMQ_XPUB_VERBOSE
 * EVENTS = ZMQ_EVENTS
 * TYPE = ZMQ_TYPE             # <<<<<<<<<<<<<<
 * LINGER = ZMQ_LINGER
 * RECONNECT_IVL = ZMQ_RECONNECT_IVL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TYPE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TYPE, __pyx_t_1) < 0) __PYX_ERR(1, 131, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":132
 * EVENTS = ZMQ_EVENTS
 * TYPE = ZMQ_TYPE
 * LINGER = ZMQ_LINGER             # <<<<<<<<<<<<<<
 * RECONNECT_IVL = ZMQ_RECONNECT_IVL
 * BACKLOG = ZMQ_BACKLOG
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_LINGER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_LINGER, __pyx_t_1) < 0) __PYX_ERR(1, 132, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":133
 * TYPE = ZMQ_TYPE
 * LINGER = ZMQ_LINGER
 * RECONNECT_IVL = ZMQ_RECONNECT_IVL             # <<<<<<<<<<<<<<
 * BACKLOG = ZMQ_BACKLOG
 * ROUTER_MANDATORY = ZMQ_ROUTER_MANDATORY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RECONNECT_IVL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECONNECT_IVL, __pyx_t_1) < 0) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":134
 * LINGER = ZMQ_LINGER
 * RECONNECT_IVL = ZMQ_RECONNECT_IVL
 * BACKLOG = ZMQ_BACKLOG             # <<<<<<<<<<<<<<
 * ROUTER_MANDATORY = ZMQ_ROUTER_MANDATORY
 * FAIL_UNROUTABLE = ZMQ_FAIL_UNROUTABLE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_BACKLOG); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_BACKLOG, __pyx_t_1) < 0) __PYX_ERR(1, 134, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":135
 * RECONNECT_IVL = ZMQ_RECONNECT_IVL
 * BACKLOG = ZMQ_BACKLOG
 * ROUTER_MANDATORY = ZMQ_ROUTER_MANDATORY             # <<<<<<<<<<<<<<
 * FAIL_UNROUTABLE = ZMQ_FAIL_UNROUTABLE
 * ROUTER_RAW = ZMQ_ROUTER_RAW
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ROUTER_MANDATORY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER_MANDATORY, __pyx_t_1) < 0) __PYX_ERR(1, 135, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":136
 * BACKLOG = ZMQ_BACKLOG
 * ROUTER_MANDATORY = ZMQ_ROUTER_MANDATORY
 * FAIL_UNROUTABLE = ZMQ_FAIL_UNROUTABLE             # <<<<<<<<<<<<<<
 * ROUTER_RAW = ZMQ_ROUTER_RAW
 * IMMEDIATE = ZMQ_IMMEDIATE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_FAIL_UNROUTABLE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_FAIL_UNROUTABLE, __pyx_t_1) < 0) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":137
 * ROUTER_MANDATORY = ZMQ_ROUTER_MANDATORY
 * FAIL_UNROUTABLE = ZMQ_FAIL_UNROUTABLE
 * ROUTER_RAW = ZMQ_ROUTER_RAW             # <<<<<<<<<<<<<<
 * IMMEDIATE = ZMQ_IMMEDIATE
 * IPV6 = ZMQ_IPV6
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ROUTER_RAW); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER_RAW, __pyx_t_1) < 0) __PYX_ERR(1, 137, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":138
 * FAIL_UNROUTABLE = ZMQ_FAIL_UNROUTABLE
 * ROUTER_RAW = ZMQ_ROUTER_RAW
 * IMMEDIATE = ZMQ_IMMEDIATE             # <<<<<<<<<<<<<<
 * IPV6 = ZMQ_IPV6
 * MECHANISM = ZMQ_MECHANISM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IMMEDIATE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IMMEDIATE, __pyx_t_1) < 0) __PYX_ERR(1, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":139
 * ROUTER_RAW = ZMQ_ROUTER_RAW
 * IMMEDIATE = ZMQ_IMMEDIATE
 * IPV6 = ZMQ_IPV6             # <<<<<<<<<<<<<<
 * MECHANISM = ZMQ_MECHANISM
 * PLAIN_SERVER = ZMQ_PLAIN_SERVER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IPV6); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IPV6, __pyx_t_1) < 0) __PYX_ERR(1, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":140
 * IMMEDIATE = ZMQ_IMMEDIATE
 * IPV6 = ZMQ_IPV6
 * MECHANISM = ZMQ_MECHANISM             # <<<<<<<<<<<<<<
 * PLAIN_SERVER = ZMQ_PLAIN_SERVER
 * CURVE_SERVER = ZMQ_CURVE_SERVER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MECHANISM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MECHANISM, __pyx_t_1) < 0) __PYX_ERR(1, 140, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":141
 * IPV6 = ZMQ_IPV6
 * MECHANISM = ZMQ_MECHANISM
 * PLAIN_SERVER = ZMQ_PLAIN_SERVER             # <<<<<<<<<<<<<<
 * CURVE_SERVER = ZMQ_CURVE_SERVER
 * PROBE_ROUTER = ZMQ_PROBE_ROUTER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PLAIN_SERVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PLAIN_SERVER, __pyx_t_1) < 0) __PYX_ERR(1, 141, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":142
 * MECHANISM = ZMQ_MECHANISM
 * PLAIN_SERVER = ZMQ_PLAIN_SERVER
 * CURVE_SERVER = ZMQ_CURVE_SERVER             # <<<<<<<<<<<<<<
 * PROBE_ROUTER = ZMQ_PROBE_ROUTER
 * REQ_RELAXED = ZMQ_REQ_RELAXED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CURVE_SERVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CURVE_SERVER, __pyx_t_1) < 0) __PYX_ERR(1, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":143
 * PLAIN_SERVER = ZMQ_PLAIN_SERVER
 * CURVE_SERVER = ZMQ_CURVE_SERVER
 * PROBE_ROUTER = ZMQ_PROBE_ROUTER             # <<<<<<<<<<<<<<
 * REQ_RELAXED = ZMQ_REQ_RELAXED
 * REQ_CORRELATE = ZMQ_REQ_CORRELATE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_PROBE_ROUTER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_PROBE_ROUTER, __pyx_t_1) < 0) __PYX_ERR(1, 143, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":144
 * CURVE_SERVER = ZMQ_CURVE_SERVER
 * PROBE_ROUTER = ZMQ_PROBE_ROUTER
 * REQ_RELAXED = ZMQ_REQ_RELAXED             # <<<<<<<<<<<<<<
 * REQ_CORRELATE = ZMQ_REQ_CORRELATE
 * CONFLATE = ZMQ_CONFLATE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_REQ_RELAXED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_REQ_RELAXED, __pyx_t_1) < 0) __PYX_ERR(1, 144, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":145
 * PROBE_ROUTER = ZMQ_PROBE_ROUTER
 * REQ_RELAXED = ZMQ_REQ_RELAXED
 * REQ_CORRELATE = ZMQ_REQ_CORRELATE             # <<<<<<<<<<<<<<
 * CONFLATE = ZMQ_CONFLATE
 * ROUTER_HANDOVER = ZMQ_ROUTER_HANDOVER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_REQ_CORRELATE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 145, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_REQ_CORRELATE, __pyx_t_1) < 0) __PYX_ERR(1, 145, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":146
 * REQ_RELAXED = ZMQ_REQ_RELAXED
 * REQ_CORRELATE = ZMQ_REQ_CORRELATE
 * CONFLATE = ZMQ_CONFLATE             # <<<<<<<<<<<<<<
 * ROUTER_HANDOVER = ZMQ_ROUTER_HANDOVER
 * TOS = ZMQ_TOS
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CONFLATE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CONFLATE, __pyx_t_1) < 0) __PYX_ERR(1, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":147
 * REQ_CORRELATE = ZMQ_REQ_CORRELATE
 * CONFLATE = ZMQ_CONFLATE
 * ROUTER_HANDOVER = ZMQ_ROUTER_HANDOVER             # <<<<<<<<<<<<<<
 * TOS = ZMQ_TOS
 * IPC_FILTER_PID = ZMQ_IPC_FILTER_PID
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_ROUTER_HANDOVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ROUTER_HANDOVER, __pyx_t_1) < 0) __PYX_ERR(1, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":148
 * CONFLATE = ZMQ_CONFLATE
 * ROUTER_HANDOVER = ZMQ_ROUTER_HANDOVER
 * TOS = ZMQ_TOS             # <<<<<<<<<<<<<<
 * IPC_FILTER_PID = ZMQ_IPC_FILTER_PID
 * IPC_FILTER_UID = ZMQ_IPC_FILTER_UID
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TOS); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TOS, __pyx_t_1) < 0) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":149
 * ROUTER_HANDOVER = ZMQ_ROUTER_HANDOVER
 * TOS = ZMQ_TOS
 * IPC_FILTER_PID = ZMQ_IPC_FILTER_PID             # <<<<<<<<<<<<<<
 * IPC_FILTER_UID = ZMQ_IPC_FILTER_UID
 * IPC_FILTER_GID = ZMQ_IPC_FILTER_GID
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IPC_FILTER_PID); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IPC_FILTER_PID, __pyx_t_1) < 0) __PYX_ERR(1, 149, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":150
 * TOS = ZMQ_TOS
 * IPC_FILTER_PID = ZMQ_IPC_FILTER_PID
 * IPC_FILTER_UID = ZMQ_IPC_FILTER_UID             # <<<<<<<<<<<<<<
 * IPC_FILTER_GID = ZMQ_IPC_FILTER_GID
 * GSSAPI_SERVER = ZMQ_GSSAPI_SERVER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IPC_FILTER_UID); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IPC_FILTER_UID, __pyx_t_1) < 0) __PYX_ERR(1, 150, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":151
 * IPC_FILTER_PID = ZMQ_IPC_FILTER_PID
 * IPC_FILTER_UID = ZMQ_IPC_FILTER_UID
 * IPC_FILTER_GID = ZMQ_IPC_FILTER_GID             # <<<<<<<<<<<<<<
 * GSSAPI_SERVER = ZMQ_GSSAPI_SERVER
 * GSSAPI_PLAINTEXT = ZMQ_GSSAPI_PLAINTEXT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_IPC_FILTER_GID); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_IPC_FILTER_GID, __pyx_t_1) < 0) __PYX_ERR(1, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":152
 * IPC_FILTER_UID = ZMQ_IPC_FILTER_UID
 * IPC_FILTER_GID = ZMQ_IPC_FILTER_GID
 * GSSAPI_SERVER = ZMQ_GSSAPI_SERVER             # <<<<<<<<<<<<<<
 * GSSAPI_PLAINTEXT = ZMQ_GSSAPI_PLAINTEXT
 * HANDSHAKE_IVL = ZMQ_HANDSHAKE_IVL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GSSAPI_SERVER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GSSAPI_SERVER, __pyx_t_1) < 0) __PYX_ERR(1, 152, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":153
 * IPC_FILTER_GID = ZMQ_IPC_FILTER_GID
 * GSSAPI_SERVER = ZMQ_GSSAPI_SERVER
 * GSSAPI_PLAINTEXT = ZMQ_GSSAPI_PLAINTEXT             # <<<<<<<<<<<<<<
 * HANDSHAKE_IVL = ZMQ_HANDSHAKE_IVL
 * XPUB_NODROP = ZMQ_XPUB_NODROP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_GSSAPI_PLAINTEXT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GSSAPI_PLAINTEXT, __pyx_t_1) < 0) __PYX_ERR(1, 153, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":154
 * GSSAPI_SERVER = ZMQ_GSSAPI_SERVER
 * GSSAPI_PLAINTEXT = ZMQ_GSSAPI_PLAINTEXT
 * HANDSHAKE_IVL = ZMQ_HANDSHAKE_IVL             # <<<<<<<<<<<<<<
 * XPUB_NODROP = ZMQ_XPUB_NODROP
 * XPUB_MANUAL = ZMQ_XPUB_MANUAL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HANDSHAKE_IVL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HANDSHAKE_IVL, __pyx_t_1) < 0) __PYX_ERR(1, 154, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":155
 * GSSAPI_PLAINTEXT = ZMQ_GSSAPI_PLAINTEXT
 * HANDSHAKE_IVL = ZMQ_HANDSHAKE_IVL
 * XPUB_NODROP = ZMQ_XPUB_NODROP             # <<<<<<<<<<<<<<
 * XPUB_MANUAL = ZMQ_XPUB_MANUAL
 * STREAM_NOTIFY = ZMQ_STREAM_NOTIFY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB_NODROP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB_NODROP, __pyx_t_1) < 0) __PYX_ERR(1, 155, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":156
 * HANDSHAKE_IVL = ZMQ_HANDSHAKE_IVL
 * XPUB_NODROP = ZMQ_XPUB_NODROP
 * XPUB_MANUAL = ZMQ_XPUB_MANUAL             # <<<<<<<<<<<<<<
 * STREAM_NOTIFY = ZMQ_STREAM_NOTIFY
 * INVERT_MATCHING = ZMQ_INVERT_MATCHING
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB_MANUAL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB_MANUAL, __pyx_t_1) < 0) __PYX_ERR(1, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":157
 * XPUB_NODROP = ZMQ_XPUB_NODROP
 * XPUB_MANUAL = ZMQ_XPUB_MANUAL
 * STREAM_NOTIFY = ZMQ_STREAM_NOTIFY             # <<<<<<<<<<<<<<
 * INVERT_MATCHING = ZMQ_INVERT_MATCHING
 * XPUB_VERBOSER = ZMQ_XPUB_VERBOSER
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_STREAM_NOTIFY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_STREAM_NOTIFY, __pyx_t_1) < 0) __PYX_ERR(1, 157, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":158
 * XPUB_MANUAL = ZMQ_XPUB_MANUAL
 * STREAM_NOTIFY = ZMQ_STREAM_NOTIFY
 * INVERT_MATCHING = ZMQ_INVERT_MATCHING             # <<<<<<<<<<<<<<
 * XPUB_VERBOSER = ZMQ_XPUB_VERBOSER
 * HEARTBEAT_IVL = ZMQ_HEARTBEAT_IVL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_INVERT_MATCHING); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INVERT_MATCHING, __pyx_t_1) < 0) __PYX_ERR(1, 158, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":159
 * STREAM_NOTIFY = ZMQ_STREAM_NOTIFY
 * INVERT_MATCHING = ZMQ_INVERT_MATCHING
 * XPUB_VERBOSER = ZMQ_XPUB_VERBOSER             # <<<<<<<<<<<<<<
 * HEARTBEAT_IVL = ZMQ_HEARTBEAT_IVL
 * HEARTBEAT_TTL = ZMQ_HEARTBEAT_TTL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_XPUB_VERBOSER); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_XPUB_VERBOSER, __pyx_t_1) < 0) __PYX_ERR(1, 159, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":160
 * INVERT_MATCHING = ZMQ_INVERT_MATCHING
 * XPUB_VERBOSER = ZMQ_XPUB_VERBOSER
 * HEARTBEAT_IVL = ZMQ_HEARTBEAT_IVL             # <<<<<<<<<<<<<<
 * HEARTBEAT_TTL = ZMQ_HEARTBEAT_TTL
 * HEARTBEAT_TIMEOUT = ZMQ_HEARTBEAT_TIMEOUT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HEARTBEAT_IVL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HEARTBEAT_IVL, __pyx_t_1) < 0) __PYX_ERR(1, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":161
 * XPUB_VERBOSER = ZMQ_XPUB_VERBOSER
 * HEARTBEAT_IVL = ZMQ_HEARTBEAT_IVL
 * HEARTBEAT_TTL = ZMQ_HEARTBEAT_TTL             # <<<<<<<<<<<<<<
 * HEARTBEAT_TIMEOUT = ZMQ_HEARTBEAT_TIMEOUT
 * CONNECT_TIMEOUT = ZMQ_CONNECT_TIMEOUT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HEARTBEAT_TTL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HEARTBEAT_TTL, __pyx_t_1) < 0) __PYX_ERR(1, 161, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":162
 * HEARTBEAT_IVL = ZMQ_HEARTBEAT_IVL
 * HEARTBEAT_TTL = ZMQ_HEARTBEAT_TTL
 * HEARTBEAT_TIMEOUT = ZMQ_HEARTBEAT_TIMEOUT             # <<<<<<<<<<<<<<
 * CONNECT_TIMEOUT = ZMQ_CONNECT_TIMEOUT
 * TCP_MAXRT = ZMQ_TCP_MAXRT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HEARTBEAT_TIMEOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HEARTBEAT_TIMEOUT, __pyx_t_1) < 0) __PYX_ERR(1, 162, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":163
 * HEARTBEAT_TTL = ZMQ_HEARTBEAT_TTL
 * HEARTBEAT_TIMEOUT = ZMQ_HEARTBEAT_TIMEOUT
 * CONNECT_TIMEOUT = ZMQ_CONNECT_TIMEOUT             # <<<<<<<<<<<<<<
 * TCP_MAXRT = ZMQ_TCP_MAXRT
 * THREAD_SAFE = ZMQ_THREAD_SAFE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_CONNECT_TIMEOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_CONNECT_TIMEOUT, __pyx_t_1) < 0) __PYX_ERR(1, 163, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":164
 * HEARTBEAT_TIMEOUT = ZMQ_HEARTBEAT_TIMEOUT
 * CONNECT_TIMEOUT = ZMQ_CONNECT_TIMEOUT
 * TCP_MAXRT = ZMQ_TCP_MAXRT             # <<<<<<<<<<<<<<
 * THREAD_SAFE = ZMQ_THREAD_SAFE
 * MULTICAST_MAXTPDU = ZMQ_MULTICAST_MAXTPDU
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_TCP_MAXRT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_TCP_MAXRT, __pyx_t_1) < 0) __PYX_ERR(1, 164, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":165
 * CONNECT_TIMEOUT = ZMQ_CONNECT_TIMEOUT
 * TCP_MAXRT = ZMQ_TCP_MAXRT
 * THREAD_SAFE = ZMQ_THREAD_SAFE             # <<<<<<<<<<<<<<
 * MULTICAST_MAXTPDU = ZMQ_MULTICAST_MAXTPDU
 * VMCI_CONNECT_TIMEOUT = ZMQ_VMCI_CONNECT_TIMEOUT
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_THREAD_SAFE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_THREAD_SAFE, __pyx_t_1) < 0) __PYX_ERR(1, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":166
 * TCP_MAXRT = ZMQ_TCP_MAXRT
 * THREAD_SAFE = ZMQ_THREAD_SAFE
 * MULTICAST_MAXTPDU = ZMQ_MULTICAST_MAXTPDU             # <<<<<<<<<<<<<<
 * VMCI_CONNECT_TIMEOUT = ZMQ_VMCI_CONNECT_TIMEOUT
 * USE_FD = ZMQ_USE_FD
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MULTICAST_MAXTPDU); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MULTICAST_MAXTPDU, __pyx_t_1) < 0) __PYX_ERR(1, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":167
 * THREAD_SAFE = ZMQ_THREAD_SAFE
 * MULTICAST_MAXTPDU = ZMQ_MULTICAST_MAXTPDU
 * VMCI_CONNECT_TIMEOUT = ZMQ_VMCI_CONNECT_TIMEOUT             # <<<<<<<<<<<<<<
 * USE_FD = ZMQ_USE_FD
 * AFFINITY = ZMQ_AFFINITY
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VMCI_CONNECT_TIMEOUT); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VMCI_CONNECT_TIMEOUT, __pyx_t_1) < 0) __PYX_ERR(1, 167, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":168
 * MULTICAST_MAXTPDU = ZMQ_MULTICAST_MAXTPDU
 * VMCI_CONNECT_TIMEOUT = ZMQ_VMCI_CONNECT_TIMEOUT
 * USE_FD = ZMQ_USE_FD             # <<<<<<<<<<<<<<
 * AFFINITY = ZMQ_AFFINITY
 * MAXMSGSIZE = ZMQ_MAXMSGSIZE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_USE_FD); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_USE_FD, __pyx_t_1) < 0) __PYX_ERR(1, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":169
 * VMCI_CONNECT_TIMEOUT = ZMQ_VMCI_CONNECT_TIMEOUT
 * USE_FD = ZMQ_USE_FD
 * AFFINITY = ZMQ_AFFINITY             # <<<<<<<<<<<<<<
 * MAXMSGSIZE = ZMQ_MAXMSGSIZE
 * HWM = ZMQ_HWM
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_AFFINITY); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AFFINITY, __pyx_t_1) < 0) __PYX_ERR(1, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":170
 * USE_FD = ZMQ_USE_FD
 * AFFINITY = ZMQ_AFFINITY
 * MAXMSGSIZE = ZMQ_MAXMSGSIZE             # <<<<<<<<<<<<<<
 * HWM = ZMQ_HWM
 * SWAP = ZMQ_SWAP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MAXMSGSIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MAXMSGSIZE, __pyx_t_1) < 0) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":171
 * AFFINITY = ZMQ_AFFINITY
 * MAXMSGSIZE = ZMQ_MAXMSGSIZE
 * HWM = ZMQ_HWM             # <<<<<<<<<<<<<<
 * SWAP = ZMQ_SWAP
 * MCAST_LOOP = ZMQ_MCAST_LOOP
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_HWM); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_HWM, __pyx_t_1) < 0) __PYX_ERR(1, 171, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":172
 * MAXMSGSIZE = ZMQ_MAXMSGSIZE
 * HWM = ZMQ_HWM
 * SWAP = ZMQ_SWAP             # <<<<<<<<<<<<<<
 * MCAST_LOOP = ZMQ_MCAST_LOOP
 * RECOVERY_IVL_MSEC = ZMQ_RECOVERY_IVL_MSEC
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SWAP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SWAP, __pyx_t_1) < 0) __PYX_ERR(1, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":173
 * HWM = ZMQ_HWM
 * SWAP = ZMQ_SWAP
 * MCAST_LOOP = ZMQ_MCAST_LOOP             # <<<<<<<<<<<<<<
 * RECOVERY_IVL_MSEC = ZMQ_RECOVERY_IVL_MSEC
 * VMCI_BUFFER_SIZE = ZMQ_VMCI_BUFFER_SIZE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MCAST_LOOP); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MCAST_LOOP, __pyx_t_1) < 0) __PYX_ERR(1, 173, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":174
 * SWAP = ZMQ_SWAP
 * MCAST_LOOP = ZMQ_MCAST_LOOP
 * RECOVERY_IVL_MSEC = ZMQ_RECOVERY_IVL_MSEC             # <<<<<<<<<<<<<<
 * VMCI_BUFFER_SIZE = ZMQ_VMCI_BUFFER_SIZE
 * VMCI_BUFFER_MIN_SIZE = ZMQ_VMCI_BUFFER_MIN_SIZE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RECOVERY_IVL_MSEC); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECOVERY_IVL_MSEC, __pyx_t_1) < 0) __PYX_ERR(1, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":175
 * MCAST_LOOP = ZMQ_MCAST_LOOP
 * RECOVERY_IVL_MSEC = ZMQ_RECOVERY_IVL_MSEC
 * VMCI_BUFFER_SIZE = ZMQ_VMCI_BUFFER_SIZE             # <<<<<<<<<<<<<<
 * VMCI_BUFFER_MIN_SIZE = ZMQ_VMCI_BUFFER_MIN_SIZE
 * VMCI_BUFFER_MAX_SIZE = ZMQ_VMCI_BUFFER_MAX_SIZE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VMCI_BUFFER_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VMCI_BUFFER_SIZE, __pyx_t_1) < 0) __PYX_ERR(1, 175, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":176
 * RECOVERY_IVL_MSEC = ZMQ_RECOVERY_IVL_MSEC
 * VMCI_BUFFER_SIZE = ZMQ_VMCI_BUFFER_SIZE
 * VMCI_BUFFER_MIN_SIZE = ZMQ_VMCI_BUFFER_MIN_SIZE             # <<<<<<<<<<<<<<
 * VMCI_BUFFER_MAX_SIZE = ZMQ_VMCI_BUFFER_MAX_SIZE
 * RATE = ZMQ_RATE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VMCI_BUFFER_MIN_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VMCI_BUFFER_MIN_SIZE, __pyx_t_1) < 0) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":177
 * VMCI_BUFFER_SIZE = ZMQ_VMCI_BUFFER_SIZE
 * VMCI_BUFFER_MIN_SIZE = ZMQ_VMCI_BUFFER_MIN_SIZE
 * VMCI_BUFFER_MAX_SIZE = ZMQ_VMCI_BUFFER_MAX_SIZE             # <<<<<<<<<<<<<<
 * RATE = ZMQ_RATE
 * RECOVERY_IVL = ZMQ_RECOVERY_IVL
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_VMCI_BUFFER_MAX_SIZE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_VMCI_BUFFER_MAX_SIZE, __pyx_t_1) < 0) __PYX_ERR(1, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":178
 * VMCI_BUFFER_MIN_SIZE = ZMQ_VMCI_BUFFER_MIN_SIZE
 * VMCI_BUFFER_MAX_SIZE = ZMQ_VMCI_BUFFER_MAX_SIZE
 * RATE = ZMQ_RATE             # <<<<<<<<<<<<<<
 * RECOVERY_IVL = ZMQ_RECOVERY_IVL
 * SNDBUF = ZMQ_SNDBUF
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RATE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RATE, __pyx_t_1) < 0) __PYX_ERR(1, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":179
 * VMCI_BUFFER_MAX_SIZE = ZMQ_VMCI_BUFFER_MAX_SIZE
 * RATE = ZMQ_RATE
 * RECOVERY_IVL = ZMQ_RECOVERY_IVL             # <<<<<<<<<<<<<<
 * SNDBUF = ZMQ_SNDBUF
 * RCVBUF = ZMQ_RCVBUF
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RECOVERY_IVL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RECOVERY_IVL, __pyx_t_1) < 0) __PYX_ERR(1, 179, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":180
 * RATE = ZMQ_RATE
 * RECOVERY_IVL = ZMQ_RECOVERY_IVL
 * SNDBUF = ZMQ_SNDBUF             # <<<<<<<<<<<<<<
 * RCVBUF = ZMQ_RCVBUF
 * RCVMORE = ZMQ_RCVMORE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SNDBUF); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SNDBUF, __pyx_t_1) < 0) __PYX_ERR(1, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":181
 * RECOVERY_IVL = ZMQ_RECOVERY_IVL
 * SNDBUF = ZMQ_SNDBUF
 * RCVBUF = ZMQ_RCVBUF             # <<<<<<<<<<<<<<
 * RCVMORE = ZMQ_RCVMORE
 * MORE = ZMQ_MORE
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RCVBUF); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RCVBUF, __pyx_t_1) < 0) __PYX_ERR(1, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":182
 * SNDBUF = ZMQ_SNDBUF
 * RCVBUF = ZMQ_RCVBUF
 * RCVMORE = ZMQ_RCVMORE             # <<<<<<<<<<<<<<
 * MORE = ZMQ_MORE
 * SRCFD = ZMQ_SRCFD
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_RCVMORE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_RCVMORE, __pyx_t_1) < 0) __PYX_ERR(1, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":183
 * RCVBUF = ZMQ_RCVBUF
 * RCVMORE = ZMQ_RCVMORE
 * MORE = ZMQ_MORE             # <<<<<<<<<<<<<<
 * SRCFD = ZMQ_SRCFD
 * SHARED = ZMQ_SHARED
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_MORE); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_MORE, __pyx_t_1) < 0) __PYX_ERR(1, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":184
 * RCVMORE = ZMQ_RCVMORE
 * MORE = ZMQ_MORE
 * SRCFD = ZMQ_SRCFD             # <<<<<<<<<<<<<<
 * SHARED = ZMQ_SHARED
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SRCFD); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SRCFD, __pyx_t_1) < 0) __PYX_ERR(1, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":185
 * MORE = ZMQ_MORE
 * SRCFD = ZMQ_SRCFD
 * SHARED = ZMQ_SHARED             # <<<<<<<<<<<<<<
 * 
 * #-----------------------------------------------------------------------------
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(ZMQ_SHARED); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 185, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_SHARED, __pyx_t_1) < 0) __PYX_ERR(1, 185, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pxi":190
 * # Symbols to export
 * #-----------------------------------------------------------------------------
 * __all__ = [             # <<<<<<<<<<<<<<
 *   "DRAFT_API",
 *   "VERSION",
 */
  __pyx_t_1 = PyList_New(180); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_DRAFT_API);
  __Pyx_GIVEREF(__pyx_n_s_DRAFT_API);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_DRAFT_API);
  __Pyx_INCREF(__pyx_n_s_VERSION);
  __Pyx_GIVEREF(__pyx_n_s_VERSION);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_VERSION);
  __Pyx_INCREF(__pyx_n_s_VERSION_MAJOR);
  __Pyx_GIVEREF(__pyx_n_s_VERSION_MAJOR);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_VERSION_MAJOR);
  __Pyx_INCREF(__pyx_n_s_VERSION_MINOR);
  __Pyx_GIVEREF(__pyx_n_s_VERSION_MINOR);
  PyList_SET_ITEM(__pyx_t_1, 3, __pyx_n_s_VERSION_MINOR);
  __Pyx_INCREF(__pyx_n_s_VERSION_PATCH);
  __Pyx_GIVEREF(__pyx_n_s_VERSION_PATCH);
  PyList_SET_ITEM(__pyx_t_1, 4, __pyx_n_s_VERSION_PATCH);
  __Pyx_INCREF(__pyx_n_s_NOBLOCK);
  __Pyx_GIVEREF(__pyx_n_s_NOBLOCK);
  PyList_SET_ITEM(__pyx_t_1, 5, __pyx_n_s_NOBLOCK);
  __Pyx_INCREF(__pyx_n_s_DONTWAIT);
  __Pyx_GIVEREF(__pyx_n_s_DONTWAIT);
  PyList_SET_ITEM(__pyx_t_1, 6, __pyx_n_s_DONTWAIT);
  __Pyx_INCREF(__pyx_n_s_POLLIN);
  __Pyx_GIVEREF(__pyx_n_s_POLLIN);
  PyList_SET_ITEM(__pyx_t_1, 7, __pyx_n_s_POLLIN);
  __Pyx_INCREF(__pyx_n_s_POLLOUT);
  __Pyx_GIVEREF(__pyx_n_s_POLLOUT);
  PyList_SET_ITEM(__pyx_t_1, 8, __pyx_n_s_POLLOUT);
  __Pyx_INCREF(__pyx_n_s_POLLERR);
  __Pyx_GIVEREF(__pyx_n_s_POLLERR);
  PyList_SET_ITEM(__pyx_t_1, 9, __pyx_n_s_POLLERR);
  __Pyx_INCREF(__pyx_n_s_POLLPRI);
  __Pyx_GIVEREF(__pyx_n_s_POLLPRI);
  PyList_SET_ITEM(__pyx_t_1, 10, __pyx_n_s_POLLPRI);
  __Pyx_INCREF(__pyx_n_s_SNDMORE);
  __Pyx_GIVEREF(__pyx_n_s_SNDMORE);
  PyList_SET_ITEM(__pyx_t_1, 11, __pyx_n_s_SNDMORE);
  __Pyx_INCREF(__pyx_n_s_STREAMER);
  __Pyx_GIVEREF(__pyx_n_s_STREAMER);
  PyList_SET_ITEM(__pyx_t_1, 12, __pyx_n_s_STREAMER);
  __Pyx_INCREF(__pyx_n_s_FORWARDER);
  __Pyx_GIVEREF(__pyx_n_s_FORWARDER);
  PyList_SET_ITEM(__pyx_t_1, 13, __pyx_n_s_FORWARDER);
  __Pyx_INCREF(__pyx_n_s_QUEUE);
  __Pyx_GIVEREF(__pyx_n_s_QUEUE);
  PyList_SET_ITEM(__pyx_t_1, 14, __pyx_n_s_QUEUE);
  __Pyx_INCREF(__pyx_n_s_IO_THREADS_DFLT);
  __Pyx_GIVEREF(__pyx_n_s_IO_THREADS_DFLT);
  PyList_SET_ITEM(__pyx_t_1, 15, __pyx_n_s_IO_THREADS_DFLT);
  __Pyx_INCREF(__pyx_n_s_MAX_SOCKETS_DFLT);
  __Pyx_GIVEREF(__pyx_n_s_MAX_SOCKETS_DFLT);
  PyList_SET_ITEM(__pyx_t_1, 16, __pyx_n_s_MAX_SOCKETS_DFLT);
  __Pyx_INCREF(__pyx_n_s_POLLITEMS_DFLT);
  __Pyx_GIVEREF(__pyx_n_s_POLLITEMS_DFLT);
  PyList_SET_ITEM(__pyx_t_1, 17, __pyx_n_s_POLLITEMS_DFLT);
  __Pyx_INCREF(__pyx_n_s_THREAD_PRIORITY_DFLT);
  __Pyx_GIVEREF(__pyx_n_s_THREAD_PRIORITY_DFLT);
  PyList_SET_ITEM(__pyx_t_1, 18, __pyx_n_s_THREAD_PRIORITY_DFLT);
  __Pyx_INCREF(__pyx_n_s_THREAD_SCHED_POLICY_DFLT);
  __Pyx_GIVEREF(__pyx_n_s_THREAD_SCHED_POLICY_DFLT);
  PyList_SET_ITEM(__pyx_t_1, 19, __pyx_n_s_THREAD_SCHED_POLICY_DFLT);
  __Pyx_INCREF(__pyx_n_s_PAIR);
  __Pyx_GIVEREF(__pyx_n_s_PAIR);
  PyList_SET_ITEM(__pyx_t_1, 20, __pyx_n_s_PAIR);
  __Pyx_INCREF(__pyx_n_s_PUB);
  __Pyx_GIVEREF(__pyx_n_s_PUB);
  PyList_SET_ITEM(__pyx_t_1, 21, __pyx_n_s_PUB);
  __Pyx_INCREF(__pyx_n_s_SUB);
  __Pyx_GIVEREF(__pyx_n_s_SUB);
  PyList_SET_ITEM(__pyx_t_1, 22, __pyx_n_s_SUB);
  __Pyx_INCREF(__pyx_n_s_REQ);
  __Pyx_GIVEREF(__pyx_n_s_REQ);
  PyList_SET_ITEM(__pyx_t_1, 23, __pyx_n_s_REQ);
  __Pyx_INCREF(__pyx_n_s_REP);
  __Pyx_GIVEREF(__pyx_n_s_REP);
  PyList_SET_ITEM(__pyx_t_1, 24, __pyx_n_s_REP);
  __Pyx_INCREF(__pyx_n_s_DEALER);
  __Pyx_GIVEREF(__pyx_n_s_DEALER);
  PyList_SET_ITEM(__pyx_t_1, 25, __pyx_n_s_DEALER);
  __Pyx_INCREF(__pyx_n_s_ROUTER);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER);
  PyList_SET_ITEM(__pyx_t_1, 26, __pyx_n_s_ROUTER);
  __Pyx_INCREF(__pyx_n_s_XREQ);
  __Pyx_GIVEREF(__pyx_n_s_XREQ);
  PyList_SET_ITEM(__pyx_t_1, 27, __pyx_n_s_XREQ);
  __Pyx_INCREF(__pyx_n_s_XREP);
  __Pyx_GIVEREF(__pyx_n_s_XREP);
  PyList_SET_ITEM(__pyx_t_1, 28, __pyx_n_s_XREP);
  __Pyx_INCREF(__pyx_n_s_PULL);
  __Pyx_GIVEREF(__pyx_n_s_PULL);
  PyList_SET_ITEM(__pyx_t_1, 29, __pyx_n_s_PULL);
  __Pyx_INCREF(__pyx_n_s_PUSH);
  __Pyx_GIVEREF(__pyx_n_s_PUSH);
  PyList_SET_ITEM(__pyx_t_1, 30, __pyx_n_s_PUSH);
  __Pyx_INCREF(__pyx_n_s_XPUB);
  __Pyx_GIVEREF(__pyx_n_s_XPUB);
  PyList_SET_ITEM(__pyx_t_1, 31, __pyx_n_s_XPUB);
  __Pyx_INCREF(__pyx_n_s_XSUB);
  __Pyx_GIVEREF(__pyx_n_s_XSUB);
  PyList_SET_ITEM(__pyx_t_1, 32, __pyx_n_s_XSUB);
  __Pyx_INCREF(__pyx_n_s_UPSTREAM);
  __Pyx_GIVEREF(__pyx_n_s_UPSTREAM);
  PyList_SET_ITEM(__pyx_t_1, 33, __pyx_n_s_UPSTREAM);
  __Pyx_INCREF(__pyx_n_s_DOWNSTREAM);
  __Pyx_GIVEREF(__pyx_n_s_DOWNSTREAM);
  PyList_SET_ITEM(__pyx_t_1, 34, __pyx_n_s_DOWNSTREAM);
  __Pyx_INCREF(__pyx_n_s_STREAM);
  __Pyx_GIVEREF(__pyx_n_s_STREAM);
  PyList_SET_ITEM(__pyx_t_1, 35, __pyx_n_s_STREAM);
  __Pyx_INCREF(__pyx_n_s_SERVER);
  __Pyx_GIVEREF(__pyx_n_s_SERVER);
  PyList_SET_ITEM(__pyx_t_1, 36, __pyx_n_s_SERVER);
  __Pyx_INCREF(__pyx_n_s_CLIENT);
  __Pyx_GIVEREF(__pyx_n_s_CLIENT);
  PyList_SET_ITEM(__pyx_t_1, 37, __pyx_n_s_CLIENT);
  __Pyx_INCREF(__pyx_n_s_RADIO);
  __Pyx_GIVEREF(__pyx_n_s_RADIO);
  PyList_SET_ITEM(__pyx_t_1, 38, __pyx_n_s_RADIO);
  __Pyx_INCREF(__pyx_n_s_DISH);
  __Pyx_GIVEREF(__pyx_n_s_DISH);
  PyList_SET_ITEM(__pyx_t_1, 39, __pyx_n_s_DISH);
  __Pyx_INCREF(__pyx_n_s_GATHER);
  __Pyx_GIVEREF(__pyx_n_s_GATHER);
  PyList_SET_ITEM(__pyx_t_1, 40, __pyx_n_s_GATHER);
  __Pyx_INCREF(__pyx_n_s_SCATTER);
  __Pyx_GIVEREF(__pyx_n_s_SCATTER);
  PyList_SET_ITEM(__pyx_t_1, 41, __pyx_n_s_SCATTER);
  __Pyx_INCREF(__pyx_n_s_DGRAM);
  __Pyx_GIVEREF(__pyx_n_s_DGRAM);
  PyList_SET_ITEM(__pyx_t_1, 42, __pyx_n_s_DGRAM);
  __Pyx_INCREF(__pyx_n_s_EVENT_CONNECTED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_CONNECTED);
  PyList_SET_ITEM(__pyx_t_1, 43, __pyx_n_s_EVENT_CONNECTED);
  __Pyx_INCREF(__pyx_n_s_EVENT_CONNECT_DELAYED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_CONNECT_DELAYED);
  PyList_SET_ITEM(__pyx_t_1, 44, __pyx_n_s_EVENT_CONNECT_DELAYED);
  __Pyx_INCREF(__pyx_n_s_EVENT_CONNECT_RETRIED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_CONNECT_RETRIED);
  PyList_SET_ITEM(__pyx_t_1, 45, __pyx_n_s_EVENT_CONNECT_RETRIED);
  __Pyx_INCREF(__pyx_n_s_EVENT_LISTENING);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_LISTENING);
  PyList_SET_ITEM(__pyx_t_1, 46, __pyx_n_s_EVENT_LISTENING);
  __Pyx_INCREF(__pyx_n_s_EVENT_BIND_FAILED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_BIND_FAILED);
  PyList_SET_ITEM(__pyx_t_1, 47, __pyx_n_s_EVENT_BIND_FAILED);
  __Pyx_INCREF(__pyx_n_s_EVENT_ACCEPTED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_ACCEPTED);
  PyList_SET_ITEM(__pyx_t_1, 48, __pyx_n_s_EVENT_ACCEPTED);
  __Pyx_INCREF(__pyx_n_s_EVENT_ACCEPT_FAILED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_ACCEPT_FAILED);
  PyList_SET_ITEM(__pyx_t_1, 49, __pyx_n_s_EVENT_ACCEPT_FAILED);
  __Pyx_INCREF(__pyx_n_s_EVENT_CLOSED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_CLOSED);
  PyList_SET_ITEM(__pyx_t_1, 50, __pyx_n_s_EVENT_CLOSED);
  __Pyx_INCREF(__pyx_n_s_EVENT_CLOSE_FAILED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_CLOSE_FAILED);
  PyList_SET_ITEM(__pyx_t_1, 51, __pyx_n_s_EVENT_CLOSE_FAILED);
  __Pyx_INCREF(__pyx_n_s_EVENT_DISCONNECTED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_DISCONNECTED);
  PyList_SET_ITEM(__pyx_t_1, 52, __pyx_n_s_EVENT_DISCONNECTED);
  __Pyx_INCREF(__pyx_n_s_EVENT_ALL);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_ALL);
  PyList_SET_ITEM(__pyx_t_1, 53, __pyx_n_s_EVENT_ALL);
  __Pyx_INCREF(__pyx_n_s_EVENT_MONITOR_STOPPED);
  __Pyx_GIVEREF(__pyx_n_s_EVENT_MONITOR_STOPPED);
  PyList_SET_ITEM(__pyx_t_1, 54, __pyx_n_s_EVENT_MONITOR_STOPPED);
  __Pyx_INCREF(__pyx_n_s_NULL);
  __Pyx_GIVEREF(__pyx_n_s_NULL);
  PyList_SET_ITEM(__pyx_t_1, 55, __pyx_n_s_NULL);
  __Pyx_INCREF(__pyx_n_s_PLAIN);
  __Pyx_GIVEREF(__pyx_n_s_PLAIN);
  PyList_SET_ITEM(__pyx_t_1, 56, __pyx_n_s_PLAIN);
  __Pyx_INCREF(__pyx_n_s_CURVE);
  __Pyx_GIVEREF(__pyx_n_s_CURVE);
  PyList_SET_ITEM(__pyx_t_1, 57, __pyx_n_s_CURVE);
  __Pyx_INCREF(__pyx_n_s_GSSAPI);
  __Pyx_GIVEREF(__pyx_n_s_GSSAPI);
  PyList_SET_ITEM(__pyx_t_1, 58, __pyx_n_s_GSSAPI);
  __Pyx_INCREF(__pyx_n_s_EAGAIN);
  __Pyx_GIVEREF(__pyx_n_s_EAGAIN);
  PyList_SET_ITEM(__pyx_t_1, 59, __pyx_n_s_EAGAIN);
  __Pyx_INCREF(__pyx_n_s_EINVAL);
  __Pyx_GIVEREF(__pyx_n_s_EINVAL);
  PyList_SET_ITEM(__pyx_t_1, 60, __pyx_n_s_EINVAL);
  __Pyx_INCREF(__pyx_n_s_EFAULT);
  __Pyx_GIVEREF(__pyx_n_s_EFAULT);
  PyList_SET_ITEM(__pyx_t_1, 61, __pyx_n_s_EFAULT);
  __Pyx_INCREF(__pyx_n_s_ENOMEM);
  __Pyx_GIVEREF(__pyx_n_s_ENOMEM);
  PyList_SET_ITEM(__pyx_t_1, 62, __pyx_n_s_ENOMEM);
  __Pyx_INCREF(__pyx_n_s_ENODEV);
  __Pyx_GIVEREF(__pyx_n_s_ENODEV);
  PyList_SET_ITEM(__pyx_t_1, 63, __pyx_n_s_ENODEV);
  __Pyx_INCREF(__pyx_n_s_EMSGSIZE);
  __Pyx_GIVEREF(__pyx_n_s_EMSGSIZE);
  PyList_SET_ITEM(__pyx_t_1, 64, __pyx_n_s_EMSGSIZE);
  __Pyx_INCREF(__pyx_n_s_EAFNOSUPPORT);
  __Pyx_GIVEREF(__pyx_n_s_EAFNOSUPPORT);
  PyList_SET_ITEM(__pyx_t_1, 65, __pyx_n_s_EAFNOSUPPORT);
  __Pyx_INCREF(__pyx_n_s_ENETUNREACH);
  __Pyx_GIVEREF(__pyx_n_s_ENETUNREACH);
  PyList_SET_ITEM(__pyx_t_1, 66, __pyx_n_s_ENETUNREACH);
  __Pyx_INCREF(__pyx_n_s_ECONNABORTED);
  __Pyx_GIVEREF(__pyx_n_s_ECONNABORTED);
  PyList_SET_ITEM(__pyx_t_1, 67, __pyx_n_s_ECONNABORTED);
  __Pyx_INCREF(__pyx_n_s_ECONNRESET);
  __Pyx_GIVEREF(__pyx_n_s_ECONNRESET);
  PyList_SET_ITEM(__pyx_t_1, 68, __pyx_n_s_ECONNRESET);
  __Pyx_INCREF(__pyx_n_s_ENOTCONN);
  __Pyx_GIVEREF(__pyx_n_s_ENOTCONN);
  PyList_SET_ITEM(__pyx_t_1, 69, __pyx_n_s_ENOTCONN);
  __Pyx_INCREF(__pyx_n_s_ETIMEDOUT);
  __Pyx_GIVEREF(__pyx_n_s_ETIMEDOUT);
  PyList_SET_ITEM(__pyx_t_1, 70, __pyx_n_s_ETIMEDOUT);
  __Pyx_INCREF(__pyx_n_s_EHOSTUNREACH);
  __Pyx_GIVEREF(__pyx_n_s_EHOSTUNREACH);
  PyList_SET_ITEM(__pyx_t_1, 71, __pyx_n_s_EHOSTUNREACH);
  __Pyx_INCREF(__pyx_n_s_ENETRESET);
  __Pyx_GIVEREF(__pyx_n_s_ENETRESET);
  PyList_SET_ITEM(__pyx_t_1, 72, __pyx_n_s_ENETRESET);
  __Pyx_INCREF(__pyx_n_s_HAUSNUMERO);
  __Pyx_GIVEREF(__pyx_n_s_HAUSNUMERO);
  PyList_SET_ITEM(__pyx_t_1, 73, __pyx_n_s_HAUSNUMERO);
  __Pyx_INCREF(__pyx_n_s_ENOTSUP);
  __Pyx_GIVEREF(__pyx_n_s_ENOTSUP);
  PyList_SET_ITEM(__pyx_t_1, 74, __pyx_n_s_ENOTSUP);
  __Pyx_INCREF(__pyx_n_s_EPROTONOSUPPORT);
  __Pyx_GIVEREF(__pyx_n_s_EPROTONOSUPPORT);
  PyList_SET_ITEM(__pyx_t_1, 75, __pyx_n_s_EPROTONOSUPPORT);
  __Pyx_INCREF(__pyx_n_s_ENOBUFS);
  __Pyx_GIVEREF(__pyx_n_s_ENOBUFS);
  PyList_SET_ITEM(__pyx_t_1, 76, __pyx_n_s_ENOBUFS);
  __Pyx_INCREF(__pyx_n_s_ENETDOWN);
  __Pyx_GIVEREF(__pyx_n_s_ENETDOWN);
  PyList_SET_ITEM(__pyx_t_1, 77, __pyx_n_s_ENETDOWN);
  __Pyx_INCREF(__pyx_n_s_EADDRINUSE);
  __Pyx_GIVEREF(__pyx_n_s_EADDRINUSE);
  PyList_SET_ITEM(__pyx_t_1, 78, __pyx_n_s_EADDRINUSE);
  __Pyx_INCREF(__pyx_n_s_EADDRNOTAVAIL);
  __Pyx_GIVEREF(__pyx_n_s_EADDRNOTAVAIL);
  PyList_SET_ITEM(__pyx_t_1, 79, __pyx_n_s_EADDRNOTAVAIL);
  __Pyx_INCREF(__pyx_n_s_ECONNREFUSED);
  __Pyx_GIVEREF(__pyx_n_s_ECONNREFUSED);
  PyList_SET_ITEM(__pyx_t_1, 80, __pyx_n_s_ECONNREFUSED);
  __Pyx_INCREF(__pyx_n_s_EINPROGRESS);
  __Pyx_GIVEREF(__pyx_n_s_EINPROGRESS);
  PyList_SET_ITEM(__pyx_t_1, 81, __pyx_n_s_EINPROGRESS);
  __Pyx_INCREF(__pyx_n_s_ENOTSOCK);
  __Pyx_GIVEREF(__pyx_n_s_ENOTSOCK);
  PyList_SET_ITEM(__pyx_t_1, 82, __pyx_n_s_ENOTSOCK);
  __Pyx_INCREF(__pyx_n_s_EFSM);
  __Pyx_GIVEREF(__pyx_n_s_EFSM);
  PyList_SET_ITEM(__pyx_t_1, 83, __pyx_n_s_EFSM);
  __Pyx_INCREF(__pyx_n_s_ENOCOMPATPROTO);
  __Pyx_GIVEREF(__pyx_n_s_ENOCOMPATPROTO);
  PyList_SET_ITEM(__pyx_t_1, 84, __pyx_n_s_ENOCOMPATPROTO);
  __Pyx_INCREF(__pyx_n_s_ETERM);
  __Pyx_GIVEREF(__pyx_n_s_ETERM);
  PyList_SET_ITEM(__pyx_t_1, 85, __pyx_n_s_ETERM);
  __Pyx_INCREF(__pyx_n_s_EMTHREAD);
  __Pyx_GIVEREF(__pyx_n_s_EMTHREAD);
  PyList_SET_ITEM(__pyx_t_1, 86, __pyx_n_s_EMTHREAD);
  __Pyx_INCREF(__pyx_n_s_IO_THREADS);
  __Pyx_GIVEREF(__pyx_n_s_IO_THREADS);
  PyList_SET_ITEM(__pyx_t_1, 87, __pyx_n_s_IO_THREADS);
  __Pyx_INCREF(__pyx_n_s_MAX_SOCKETS);
  __Pyx_GIVEREF(__pyx_n_s_MAX_SOCKETS);
  PyList_SET_ITEM(__pyx_t_1, 88, __pyx_n_s_MAX_SOCKETS);
  __Pyx_INCREF(__pyx_n_s_SOCKET_LIMIT);
  __Pyx_GIVEREF(__pyx_n_s_SOCKET_LIMIT);
  PyList_SET_ITEM(__pyx_t_1, 89, __pyx_n_s_SOCKET_LIMIT);
  __Pyx_INCREF(__pyx_n_s_THREAD_PRIORITY);
  __Pyx_GIVEREF(__pyx_n_s_THREAD_PRIORITY);
  PyList_SET_ITEM(__pyx_t_1, 90, __pyx_n_s_THREAD_PRIORITY);
  __Pyx_INCREF(__pyx_n_s_THREAD_SCHED_POLICY);
  __Pyx_GIVEREF(__pyx_n_s_THREAD_SCHED_POLICY);
  PyList_SET_ITEM(__pyx_t_1, 91, __pyx_n_s_THREAD_SCHED_POLICY);
  __Pyx_INCREF(__pyx_n_s_BLOCKY);
  __Pyx_GIVEREF(__pyx_n_s_BLOCKY);
  PyList_SET_ITEM(__pyx_t_1, 92, __pyx_n_s_BLOCKY);
  __Pyx_INCREF(__pyx_n_s_IDENTITY);
  __Pyx_GIVEREF(__pyx_n_s_IDENTITY);
  PyList_SET_ITEM(__pyx_t_1, 93, __pyx_n_s_IDENTITY);
  __Pyx_INCREF(__pyx_n_s_SUBSCRIBE);
  __Pyx_GIVEREF(__pyx_n_s_SUBSCRIBE);
  PyList_SET_ITEM(__pyx_t_1, 94, __pyx_n_s_SUBSCRIBE);
  __Pyx_INCREF(__pyx_n_s_UNSUBSCRIBE);
  __Pyx_GIVEREF(__pyx_n_s_UNSUBSCRIBE);
  PyList_SET_ITEM(__pyx_t_1, 95, __pyx_n_s_UNSUBSCRIBE);
  __Pyx_INCREF(__pyx_n_s_LAST_ENDPOINT);
  __Pyx_GIVEREF(__pyx_n_s_LAST_ENDPOINT);
  PyList_SET_ITEM(__pyx_t_1, 96, __pyx_n_s_LAST_ENDPOINT);
  __Pyx_INCREF(__pyx_n_s_TCP_ACCEPT_FILTER);
  __Pyx_GIVEREF(__pyx_n_s_TCP_ACCEPT_FILTER);
  PyList_SET_ITEM(__pyx_t_1, 97, __pyx_n_s_TCP_ACCEPT_FILTER);
  __Pyx_INCREF(__pyx_n_s_PLAIN_USERNAME);
  __Pyx_GIVEREF(__pyx_n_s_PLAIN_USERNAME);
  PyList_SET_ITEM(__pyx_t_1, 98, __pyx_n_s_PLAIN_USERNAME);
  __Pyx_INCREF(__pyx_n_s_PLAIN_PASSWORD);
  __Pyx_GIVEREF(__pyx_n_s_PLAIN_PASSWORD);
  PyList_SET_ITEM(__pyx_t_1, 99, __pyx_n_s_PLAIN_PASSWORD);
  __Pyx_INCREF(__pyx_n_s_CURVE_PUBLICKEY);
  __Pyx_GIVEREF(__pyx_n_s_CURVE_PUBLICKEY);
  PyList_SET_ITEM(__pyx_t_1, 100, __pyx_n_s_CURVE_PUBLICKEY);
  __Pyx_INCREF(__pyx_n_s_CURVE_SECRETKEY);
  __Pyx_GIVEREF(__pyx_n_s_CURVE_SECRETKEY);
  PyList_SET_ITEM(__pyx_t_1, 101, __pyx_n_s_CURVE_SECRETKEY);
  __Pyx_INCREF(__pyx_n_s_CURVE_SERVERKEY);
  __Pyx_GIVEREF(__pyx_n_s_CURVE_SERVERKEY);
  PyList_SET_ITEM(__pyx_t_1, 102, __pyx_n_s_CURVE_SERVERKEY);
  __Pyx_INCREF(__pyx_n_s_ZAP_DOMAIN);
  __Pyx_GIVEREF(__pyx_n_s_ZAP_DOMAIN);
  PyList_SET_ITEM(__pyx_t_1, 103, __pyx_n_s_ZAP_DOMAIN);
  __Pyx_INCREF(__pyx_n_s_CONNECT_RID);
  __Pyx_GIVEREF(__pyx_n_s_CONNECT_RID);
  PyList_SET_ITEM(__pyx_t_1, 104, __pyx_n_s_CONNECT_RID);
  __Pyx_INCREF(__pyx_n_s_GSSAPI_PRINCIPAL);
  __Pyx_GIVEREF(__pyx_n_s_GSSAPI_PRINCIPAL);
  PyList_SET_ITEM(__pyx_t_1, 105, __pyx_n_s_GSSAPI_PRINCIPAL);
  __Pyx_INCREF(__pyx_n_s_GSSAPI_SERVICE_PRINCIPAL);
  __Pyx_GIVEREF(__pyx_n_s_GSSAPI_SERVICE_PRINCIPAL);
  PyList_SET_ITEM(__pyx_t_1, 106, __pyx_n_s_GSSAPI_SERVICE_PRINCIPAL);
  __Pyx_INCREF(__pyx_n_s_SOCKS_PROXY);
  __Pyx_GIVEREF(__pyx_n_s_SOCKS_PROXY);
  PyList_SET_ITEM(__pyx_t_1, 107, __pyx_n_s_SOCKS_PROXY);
  __Pyx_INCREF(__pyx_n_s_XPUB_WELCOME_MSG);
  __Pyx_GIVEREF(__pyx_n_s_XPUB_WELCOME_MSG);
  PyList_SET_ITEM(__pyx_t_1, 108, __pyx_n_s_XPUB_WELCOME_MSG);
  __Pyx_INCREF(__pyx_n_s_FD);
  __Pyx_GIVEREF(__pyx_n_s_FD);
  PyList_SET_ITEM(__pyx_t_1, 109, __pyx_n_s_FD);
  __Pyx_INCREF(__pyx_n_s_RECONNECT_IVL_MAX);
  __Pyx_GIVEREF(__pyx_n_s_RECONNECT_IVL_MAX);
  PyList_SET_ITEM(__pyx_t_1, 110, __pyx_n_s_RECONNECT_IVL_MAX);
  __Pyx_INCREF(__pyx_n_s_SNDTIMEO);
  __Pyx_GIVEREF(__pyx_n_s_SNDTIMEO);
  PyList_SET_ITEM(__pyx_t_1, 111, __pyx_n_s_SNDTIMEO);
  __Pyx_INCREF(__pyx_n_s_RCVTIMEO);
  __Pyx_GIVEREF(__pyx_n_s_RCVTIMEO);
  PyList_SET_ITEM(__pyx_t_1, 112, __pyx_n_s_RCVTIMEO);
  __Pyx_INCREF(__pyx_n_s_SNDHWM);
  __Pyx_GIVEREF(__pyx_n_s_SNDHWM);
  PyList_SET_ITEM(__pyx_t_1, 113, __pyx_n_s_SNDHWM);
  __Pyx_INCREF(__pyx_n_s_RCVHWM);
  __Pyx_GIVEREF(__pyx_n_s_RCVHWM);
  PyList_SET_ITEM(__pyx_t_1, 114, __pyx_n_s_RCVHWM);
  __Pyx_INCREF(__pyx_n_s_MULTICAST_HOPS);
  __Pyx_GIVEREF(__pyx_n_s_MULTICAST_HOPS);
  PyList_SET_ITEM(__pyx_t_1, 115, __pyx_n_s_MULTICAST_HOPS);
  __Pyx_INCREF(__pyx_n_s_IPV4ONLY);
  __Pyx_GIVEREF(__pyx_n_s_IPV4ONLY);
  PyList_SET_ITEM(__pyx_t_1, 116, __pyx_n_s_IPV4ONLY);
  __Pyx_INCREF(__pyx_n_s_ROUTER_BEHAVIOR);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER_BEHAVIOR);
  PyList_SET_ITEM(__pyx_t_1, 117, __pyx_n_s_ROUTER_BEHAVIOR);
  __Pyx_INCREF(__pyx_n_s_TCP_KEEPALIVE);
  __Pyx_GIVEREF(__pyx_n_s_TCP_KEEPALIVE);
  PyList_SET_ITEM(__pyx_t_1, 118, __pyx_n_s_TCP_KEEPALIVE);
  __Pyx_INCREF(__pyx_n_s_TCP_KEEPALIVE_CNT);
  __Pyx_GIVEREF(__pyx_n_s_TCP_KEEPALIVE_CNT);
  PyList_SET_ITEM(__pyx_t_1, 119, __pyx_n_s_TCP_KEEPALIVE_CNT);
  __Pyx_INCREF(__pyx_n_s_TCP_KEEPALIVE_IDLE);
  __Pyx_GIVEREF(__pyx_n_s_TCP_KEEPALIVE_IDLE);
  PyList_SET_ITEM(__pyx_t_1, 120, __pyx_n_s_TCP_KEEPALIVE_IDLE);
  __Pyx_INCREF(__pyx_n_s_TCP_KEEPALIVE_INTVL);
  __Pyx_GIVEREF(__pyx_n_s_TCP_KEEPALIVE_INTVL);
  PyList_SET_ITEM(__pyx_t_1, 121, __pyx_n_s_TCP_KEEPALIVE_INTVL);
  __Pyx_INCREF(__pyx_n_s_DELAY_ATTACH_ON_CONNECT);
  __Pyx_GIVEREF(__pyx_n_s_DELAY_ATTACH_ON_CONNECT);
  PyList_SET_ITEM(__pyx_t_1, 122, __pyx_n_s_DELAY_ATTACH_ON_CONNECT);
  __Pyx_INCREF(__pyx_n_s_XPUB_VERBOSE);
  __Pyx_GIVEREF(__pyx_n_s_XPUB_VERBOSE);
  PyList_SET_ITEM(__pyx_t_1, 123, __pyx_n_s_XPUB_VERBOSE);
  __Pyx_INCREF(__pyx_n_s_EVENTS);
  __Pyx_GIVEREF(__pyx_n_s_EVENTS);
  PyList_SET_ITEM(__pyx_t_1, 124, __pyx_n_s_EVENTS);
  __Pyx_INCREF(__pyx_n_s_TYPE);
  __Pyx_GIVEREF(__pyx_n_s_TYPE);
  PyList_SET_ITEM(__pyx_t_1, 125, __pyx_n_s_TYPE);
  __Pyx_INCREF(__pyx_n_s_LINGER);
  __Pyx_GIVEREF(__pyx_n_s_LINGER);
  PyList_SET_ITEM(__pyx_t_1, 126, __pyx_n_s_LINGER);
  __Pyx_INCREF(__pyx_n_s_RECONNECT_IVL);
  __Pyx_GIVEREF(__pyx_n_s_RECONNECT_IVL);
  PyList_SET_ITEM(__pyx_t_1, 127, __pyx_n_s_RECONNECT_IVL);
  __Pyx_INCREF(__pyx_n_s_BACKLOG);
  __Pyx_GIVEREF(__pyx_n_s_BACKLOG);
  PyList_SET_ITEM(__pyx_t_1, 128, __pyx_n_s_BACKLOG);
  __Pyx_INCREF(__pyx_n_s_ROUTER_MANDATORY);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER_MANDATORY);
  PyList_SET_ITEM(__pyx_t_1, 129, __pyx_n_s_ROUTER_MANDATORY);
  __Pyx_INCREF(__pyx_n_s_FAIL_UNROUTABLE);
  __Pyx_GIVEREF(__pyx_n_s_FAIL_UNROUTABLE);
  PyList_SET_ITEM(__pyx_t_1, 130, __pyx_n_s_FAIL_UNROUTABLE);
  __Pyx_INCREF(__pyx_n_s_ROUTER_RAW);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER_RAW);
  PyList_SET_ITEM(__pyx_t_1, 131, __pyx_n_s_ROUTER_RAW);
  __Pyx_INCREF(__pyx_n_s_IMMEDIATE);
  __Pyx_GIVEREF(__pyx_n_s_IMMEDIATE);
  PyList_SET_ITEM(__pyx_t_1, 132, __pyx_n_s_IMMEDIATE);
  __Pyx_INCREF(__pyx_n_s_IPV6);
  __Pyx_GIVEREF(__pyx_n_s_IPV6);
  PyList_SET_ITEM(__pyx_t_1, 133, __pyx_n_s_IPV6);
  __Pyx_INCREF(__pyx_n_s_MECHANISM);
  __Pyx_GIVEREF(__pyx_n_s_MECHANISM);
  PyList_SET_ITEM(__pyx_t_1, 134, __pyx_n_s_MECHANISM);
  __Pyx_INCREF(__pyx_n_s_PLAIN_SERVER);
  __Pyx_GIVEREF(__pyx_n_s_PLAIN_SERVER);
  PyList_SET_ITEM(__pyx_t_1, 135, __pyx_n_s_PLAIN_SERVER);
  __Pyx_INCREF(__pyx_n_s_CURVE_SERVER);
  __Pyx_GIVEREF(__pyx_n_s_CURVE_SERVER);
  PyList_SET_ITEM(__pyx_t_1, 136, __pyx_n_s_CURVE_SERVER);
  __Pyx_INCREF(__pyx_n_s_PROBE_ROUTER);
  __Pyx_GIVEREF(__pyx_n_s_PROBE_ROUTER);
  PyList_SET_ITEM(__pyx_t_1, 137, __pyx_n_s_PROBE_ROUTER);
  __Pyx_INCREF(__pyx_n_s_REQ_RELAXED);
  __Pyx_GIVEREF(__pyx_n_s_REQ_RELAXED);
  PyList_SET_ITEM(__pyx_t_1, 138, __pyx_n_s_REQ_RELAXED);
  __Pyx_INCREF(__pyx_n_s_REQ_CORRELATE);
  __Pyx_GIVEREF(__pyx_n_s_REQ_CORRELATE);
  PyList_SET_ITEM(__pyx_t_1, 139, __pyx_n_s_REQ_CORRELATE);
  __Pyx_INCREF(__pyx_n_s_CONFLATE);
  __Pyx_GIVEREF(__pyx_n_s_CONFLATE);
  PyList_SET_ITEM(__pyx_t_1, 140, __pyx_n_s_CONFLATE);
  __Pyx_INCREF(__pyx_n_s_ROUTER_HANDOVER);
  __Pyx_GIVEREF(__pyx_n_s_ROUTER_HANDOVER);
  PyList_SET_ITEM(__pyx_t_1, 141, __pyx_n_s_ROUTER_HANDOVER);
  __Pyx_INCREF(__pyx_n_s_TOS);
  __Pyx_GIVEREF(__pyx_n_s_TOS);
  PyList_SET_ITEM(__pyx_t_1, 142, __pyx_n_s_TOS);
  __Pyx_INCREF(__pyx_n_s_IPC_FILTER_PID);
  __Pyx_GIVEREF(__pyx_n_s_IPC_FILTER_PID);
  PyList_SET_ITEM(__pyx_t_1, 143, __pyx_n_s_IPC_FILTER_PID);
  __Pyx_INCREF(__pyx_n_s_IPC_FILTER_UID);
  __Pyx_GIVEREF(__pyx_n_s_IPC_FILTER_UID);
  PyList_SET_ITEM(__pyx_t_1, 144, __pyx_n_s_IPC_FILTER_UID);
  __Pyx_INCREF(__pyx_n_s_IPC_FILTER_GID);
  __Pyx_GIVEREF(__pyx_n_s_IPC_FILTER_GID);
  PyList_SET_ITEM(__pyx_t_1, 145, __pyx_n_s_IPC_FILTER_GID);
  __Pyx_INCREF(__pyx_n_s_GSSAPI_SERVER);
  __Pyx_GIVEREF(__pyx_n_s_GSSAPI_SERVER);
  PyList_SET_ITEM(__pyx_t_1, 146, __pyx_n_s_GSSAPI_SERVER);
  __Pyx_INCREF(__pyx_n_s_GSSAPI_PLAINTEXT);
  __Pyx_GIVEREF(__pyx_n_s_GSSAPI_PLAINTEXT);
  PyList_SET_ITEM(__pyx_t_1, 147, __pyx_n_s_GSSAPI_PLAINTEXT);
  __Pyx_INCREF(__pyx_n_s_HANDSHAKE_IVL);
  __Pyx_GIVEREF(__pyx_n_s_HANDSHAKE_IVL);
  PyList_SET_ITEM(__pyx_t_1, 148, __pyx_n_s_HANDSHAKE_IVL);
  __Pyx_INCREF(__pyx_n_s_XPUB_NODROP);
  __Pyx_GIVEREF(__pyx_n_s_XPUB_NODROP);
  PyList_SET_ITEM(__pyx_t_1, 149, __pyx_n_s_XPUB_NODROP);
  __Pyx_INCREF(__pyx_n_s_XPUB_MANUAL);
  __Pyx_GIVEREF(__pyx_n_s_XPUB_MANUAL);
  PyList_SET_ITEM(__pyx_t_1, 150, __pyx_n_s_XPUB_MANUAL);
  __Pyx_INCREF(__pyx_n_s_STREAM_NOTIFY);
  __Pyx_GIVEREF(__pyx_n_s_STREAM_NOTIFY);
  PyList_SET_ITEM(__pyx_t_1, 151, __pyx_n_s_STREAM_NOTIFY);
  __Pyx_INCREF(__pyx_n_s_INVERT_MATCHING);
  __Pyx_GIVEREF(__pyx_n_s_INVERT_MATCHING);
  PyList_SET_ITEM(__pyx_t_1, 152, __pyx_n_s_INVERT_MATCHING);
  __Pyx_INCREF(__pyx_n_s_XPUB_VERBOSER);
  __Pyx_GIVEREF(__pyx_n_s_XPUB_VERBOSER);
  PyList_SET_ITEM(__pyx_t_1, 153, __pyx_n_s_XPUB_VERBOSER);
  __Pyx_INCREF(__pyx_n_s_HEARTBEAT_IVL);
  __Pyx_GIVEREF(__pyx_n_s_HEARTBEAT_IVL);
  PyList_SET_ITEM(__pyx_t_1, 154, __pyx_n_s_HEARTBEAT_IVL);
  __Pyx_INCREF(__pyx_n_s_HEARTBEAT_TTL);
  __Pyx_GIVEREF(__pyx_n_s_HEARTBEAT_TTL);
  PyList_SET_ITEM(__pyx_t_1, 155, __pyx_n_s_HEARTBEAT_TTL);
  __Pyx_INCREF(__pyx_n_s_HEARTBEAT_TIMEOUT);
  __Pyx_GIVEREF(__pyx_n_s_HEARTBEAT_TIMEOUT);
  PyList_SET_ITEM(__pyx_t_1, 156, __pyx_n_s_HEARTBEAT_TIMEOUT);
  __Pyx_INCREF(__pyx_n_s_CONNECT_TIMEOUT);
  __Pyx_GIVEREF(__pyx_n_s_CONNECT_TIMEOUT);
  PyList_SET_ITEM(__pyx_t_1, 157, __pyx_n_s_CONNECT_TIMEOUT);
  __Pyx_INCREF(__pyx_n_s_TCP_MAXRT);
  __Pyx_GIVEREF(__pyx_n_s_TCP_MAXRT);
  PyList_SET_ITEM(__pyx_t_1, 158, __pyx_n_s_TCP_MAXRT);
  __Pyx_INCREF(__pyx_n_s_THREAD_SAFE);
  __Pyx_GIVEREF(__pyx_n_s_THREAD_SAFE);
  PyList_SET_ITEM(__pyx_t_1, 159, __pyx_n_s_THREAD_SAFE);
  __Pyx_INCREF(__pyx_n_s_MULTICAST_MAXTPDU);
  __Pyx_GIVEREF(__pyx_n_s_MULTICAST_MAXTPDU);
  PyList_SET_ITEM(__pyx_t_1, 160, __pyx_n_s_MULTICAST_MAXTPDU);
  __Pyx_INCREF(__pyx_n_s_VMCI_CONNECT_TIMEOUT);
  __Pyx_GIVEREF(__pyx_n_s_VMCI_CONNECT_TIMEOUT);
  PyList_SET_ITEM(__pyx_t_1, 161, __pyx_n_s_VMCI_CONNECT_TIMEOUT);
  __Pyx_INCREF(__pyx_n_s_USE_FD);
  __Pyx_GIVEREF(__pyx_n_s_USE_FD);
  PyList_SET_ITEM(__pyx_t_1, 162, __pyx_n_s_USE_FD);
  __Pyx_INCREF(__pyx_n_s_AFFINITY);
  __Pyx_GIVEREF(__pyx_n_s_AFFINITY);
  PyList_SET_ITEM(__pyx_t_1, 163, __pyx_n_s_AFFINITY);
  __Pyx_INCREF(__pyx_n_s_MAXMSGSIZE);
  __Pyx_GIVEREF(__pyx_n_s_MAXMSGSIZE);
  PyList_SET_ITEM(__pyx_t_1, 164, __pyx_n_s_MAXMSGSIZE);
  __Pyx_INCREF(__pyx_n_s_HWM);
  __Pyx_GIVEREF(__pyx_n_s_HWM);
  PyList_SET_ITEM(__pyx_t_1, 165, __pyx_n_s_HWM);
  __Pyx_INCREF(__pyx_n_s_SWAP);
  __Pyx_GIVEREF(__pyx_n_s_SWAP);
  PyList_SET_ITEM(__pyx_t_1, 166, __pyx_n_s_SWAP);
  __Pyx_INCREF(__pyx_n_s_MCAST_LOOP);
  __Pyx_GIVEREF(__pyx_n_s_MCAST_LOOP);
  PyList_SET_ITEM(__pyx_t_1, 167, __pyx_n_s_MCAST_LOOP);
  __Pyx_INCREF(__pyx_n_s_RECOVERY_IVL_MSEC);
  __Pyx_GIVEREF(__pyx_n_s_RECOVERY_IVL_MSEC);
  PyList_SET_ITEM(__pyx_t_1, 168, __pyx_n_s_RECOVERY_IVL_MSEC);
  __Pyx_INCREF(__pyx_n_s_VMCI_BUFFER_SIZE);
  __Pyx_GIVEREF(__pyx_n_s_VMCI_BUFFER_SIZE);
  PyList_SET_ITEM(__pyx_t_1, 169, __pyx_n_s_VMCI_BUFFER_SIZE);
  __Pyx_INCREF(__pyx_n_s_VMCI_BUFFER_MIN_SIZE);
  __Pyx_GIVEREF(__pyx_n_s_VMCI_BUFFER_MIN_SIZE);
  PyList_SET_ITEM(__pyx_t_1, 170, __pyx_n_s_VMCI_BUFFER_MIN_SIZE);
  __Pyx_INCREF(__pyx_n_s_VMCI_BUFFER_MAX_SIZE);
  __Pyx_GIVEREF(__pyx_n_s_VMCI_BUFFER_MAX_SIZE);
  PyList_SET_ITEM(__pyx_t_1, 171, __pyx_n_s_VMCI_BUFFER_MAX_SIZE);
  __Pyx_INCREF(__pyx_n_s_RATE);
  __Pyx_GIVEREF(__pyx_n_s_RATE);
  PyList_SET_ITEM(__pyx_t_1, 172, __pyx_n_s_RATE);
  __Pyx_INCREF(__pyx_n_s_RECOVERY_IVL);
  __Pyx_GIVEREF(__pyx_n_s_RECOVERY_IVL);
  PyList_SET_ITEM(__pyx_t_1, 173, __pyx_n_s_RECOVERY_IVL);
  __Pyx_INCREF(__pyx_n_s_SNDBUF);
  __Pyx_GIVEREF(__pyx_n_s_SNDBUF);
  PyList_SET_ITEM(__pyx_t_1, 174, __pyx_n_s_SNDBUF);
  __Pyx_INCREF(__pyx_n_s_RCVBUF);
  __Pyx_GIVEREF(__pyx_n_s_RCVBUF);
  PyList_SET_ITEM(__pyx_t_1, 175, __pyx_n_s_RCVBUF);
  __Pyx_INCREF(__pyx_n_s_RCVMORE);
  __Pyx_GIVEREF(__pyx_n_s_RCVMORE);
  PyList_SET_ITEM(__pyx_t_1, 176, __pyx_n_s_RCVMORE);
  __Pyx_INCREF(__pyx_n_s_MORE);
  __Pyx_GIVEREF(__pyx_n_s_MORE);
  PyList_SET_ITEM(__pyx_t_1, 177, __pyx_n_s_MORE);
  __Pyx_INCREF(__pyx_n_s_SRCFD);
  __Pyx_GIVEREF(__pyx_n_s_SRCFD);
  PyList_SET_ITEM(__pyx_t_1, 178, __pyx_n_s_SRCFD);
  __Pyx_INCREF(__pyx_n_s_SHARED);
  __Pyx_GIVEREF(__pyx_n_s_SHARED);
  PyList_SET_ITEM(__pyx_t_1, 179, __pyx_n_s_SHARED);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_all, __pyx_t_1) < 0) __PYX_ERR(1, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "zmq/backend/cython/constants.pyx":1
 * """0MQ Constants."""             # <<<<<<<<<<<<<<
 * 
 * #
 */
  __pyx_t_1 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init zmq.backend.cython.constants", 0, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init zmq.backend.cython.constants");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* Globals */
static PyObject* __Pyx_Globals(void) {
    Py_ssize_t i;
    PyObject *names;
    PyObject *globals = __pyx_d;
    Py_INCREF(globals);
    names = PyObject_Dir(__pyx_m);
    if (!names)
        goto bad;
    for (i = PyList_GET_SIZE(names)-1; i >= 0; i--) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject* name = PySequence_ITEM(names, i);
        if (!name)
            goto bad;
#else
        PyObject* name = PyList_GET_ITEM(names, i);
#endif
        if (!PyDict_Contains(globals, name)) {
            PyObject* value = __Pyx_GetAttr(__pyx_m, name);
            if (!value) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                goto bad;
            }
            if (PyDict_SetItem(globals, name, value) < 0) {
#if CYTHON_COMPILING_IN_PYPY
                Py_DECREF(name);
#endif
                Py_DECREF(value);
                goto bad;
            }
        }
#if CYTHON_COMPILING_IN_PYPY
        Py_DECREF(name);
#endif
    }
    Py_DECREF(names);
    return globals;
bad:
    Py_XDECREF(names);
    Py_XDECREF(globals);
    return NULL;
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
      use_cline = __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback);
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (PyObject_Not(use_cline) != 0) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
